import torch
import torch.nn as nn
import numpy as np
from math import sqrt
#from utils.masking import TriangularCausalMask, ProbMask
#from reformer_pytorch import LSHSelfAttention
from einops import rearrange


class TriangularCausalMask():
    def __init__(self, B, L, device="cpu"):
        mask_shape = [B, 1, L, L]
        with torch.no_grad():
            self._mask = torch.triu(torch.ones(mask_shape, dtype=torch.bool), diagonal=1).to(device)

    @property
    def mask(self):
        return self._mask


class ProbMask():
    def __init__(self, B, H, L, index, scores, device="cpu"):
        _mask = torch.ones(L, scores.shape[-1], dtype=torch.bool).to(device).triu(1)
        _mask_ex = _mask[None, None, :].expand(B, H, L, scores.shape[-1])
        indicator = _mask_ex[torch.arange(B)[:, None, None],
                    torch.arange(H)[None, :, None],
                    index, :].to(device)
        self._mask = indicator.view(scores.shape).to(device)

    @property
    def mask(self):
        return self._mask

# Code implementation from https://github.com/thuml/Flowformer
class FlowAttention(nn.Module):
    def __init__(self, attention_dropout=0.1):
        super(FlowAttention, self).__init__()
        self.dropout = nn.Dropout(attention_dropout)

    def kernel_method(self, x):
        return torch.sigmoid(x)

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        queries = queries.transpose(1, 2)
        keys = keys.transpose(1, 2)
        values = values.transpose(1, 2)
        # kernel
        queries = self.kernel_method(queries)
        keys = self.kernel_method(keys)
        # incoming and outgoing
        normalizer_row = 1.0 / (torch.einsum("nhld,nhd->nhl", queries + 1e-6, keys.sum(dim=2) + 1e-6))
        normalizer_col = 1.0 / (torch.einsum("nhsd,nhd->nhs", keys + 1e-6, queries.sum(dim=2) + 1e-6))
        # reweighting
        normalizer_row_refine = (
            torch.einsum("nhld,nhd->nhl", queries + 1e-6, (keys * normalizer_col[:, :, :, None]).sum(dim=2) + 1e-6))
        normalizer_col_refine = (
            torch.einsum("nhsd,nhd->nhs", keys + 1e-6, (queries * normalizer_row[:, :, :, None]).sum(dim=2) + 1e-6))
        # competition and allocation
        normalizer_row_refine = torch.sigmoid(
            normalizer_row_refine * (float(queries.shape[2]) / float(keys.shape[2])))
        normalizer_col_refine = torch.softmax(normalizer_col_refine, dim=-1) * keys.shape[2]  # B h L vis
        # multiply
        kv = keys.transpose(-2, -1) @ (values * normalizer_col_refine[:, :, :, None])
        x = (((queries @ kv) * normalizer_row[:, :, :, None]) * normalizer_row_refine[:, :, :, None]).transpose(1,
                                                                                                                2).contiguous()
        return x, None


# Code implementation from https://github.com/shreyansh26/FlashAttention-PyTorch
class FlashAttention(nn.Module):
    def __init__(self, mask_flag=True, factor=5, scale=None, attention_dropout=0.1, output_attention=False):
        super(FlashAttention, self).__init__()
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)

    def flash_attention_forward(self, Q, K, V, mask=None):
        BLOCK_SIZE = 32
        NEG_INF = -1e10  # -infinity
        EPSILON = 1e-10
        # mask = torch.randint(0, 2, (128, 8)).to(device='cuda')
        O = torch.zeros_like(Q, requires_grad=True)
        l = torch.zeros(Q.shape[:-1])[..., None]
        m = torch.ones(Q.shape[:-1])[..., None] * NEG_INF

        O = O.to(device='cuda')
        l = l.to(device='cuda')
        m = m.to(device='cuda')

        Q_BLOCK_SIZE = min(BLOCK_SIZE, Q.shape[-1])
        KV_BLOCK_SIZE = BLOCK_SIZE

        Q_BLOCKS = torch.split(Q, Q_BLOCK_SIZE, dim=2)
        K_BLOCKS = torch.split(K, KV_BLOCK_SIZE, dim=2)
        V_BLOCKS = torch.split(V, KV_BLOCK_SIZE, dim=2)
        if mask is not None:
            mask_BLOCKS = list(torch.split(mask, KV_BLOCK_SIZE, dim=1))

        Tr = len(Q_BLOCKS)
        Tc = len(K_BLOCKS)

        O_BLOCKS = list(torch.split(O, Q_BLOCK_SIZE, dim=2))
        l_BLOCKS = list(torch.split(l, Q_BLOCK_SIZE, dim=2))
        m_BLOCKS = list(torch.split(m, Q_BLOCK_SIZE, dim=2))

        for j in range(Tc):
            Kj = K_BLOCKS[j]
            Vj = V_BLOCKS[j]
            if mask is not None:
                maskj = mask_BLOCKS[j]

            for i in range(Tr):
                Qi = Q_BLOCKS[i]
                Oi = O_BLOCKS[i]
                li = l_BLOCKS[i]
                mi = m_BLOCKS[i]

                scale = 1 / np.sqrt(Q.shape[-1])
                Qi_scaled = Qi * scale

                S_ij = torch.einsum('... i d, ... j d -> ... i j', Qi_scaled, Kj)
                if mask is not None:
                    # Masking
                    maskj_temp = rearrange(maskj, 'b j -> b 1 1 j')
                    S_ij = torch.where(maskj_temp > 0, S_ij, NEG_INF)

                m_block_ij, _ = torch.max(S_ij, dim=-1, keepdims=True)
                P_ij = torch.exp(S_ij - m_block_ij)
                if mask is not None:
                    # Masking
                    P_ij = torch.where(maskj_temp > 0, P_ij, 0.)

                l_block_ij = torch.sum(P_ij, dim=-1, keepdims=True) + EPSILON

                P_ij_Vj = torch.einsum('... i j, ... j d -> ... i d', P_ij, Vj)

                mi_new = torch.maximum(m_block_ij, mi)
                li_new = torch.exp(mi - mi_new) * li + torch.exp(m_block_ij - mi_new) * l_block_ij

                O_BLOCKS[i] = (li / li_new) * torch.exp(mi - mi_new) * Oi + (
                        torch.exp(m_block_ij - mi_new) / li_new) * P_ij_Vj
                l_BLOCKS[i] = li_new
                m_BLOCKS[i] = mi_new

        O = torch.cat(O_BLOCKS, dim=2)
        l = torch.cat(l_BLOCKS, dim=2)
        m = torch.cat(m_BLOCKS, dim=2)
        return O, l, m

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        res = \
        self.flash_attention_forward(queries.permute(0, 2, 1, 3), keys.permute(0, 2, 1, 3), values.permute(0, 2, 1, 3),
                                     attn_mask)[0]
        return res.permute(0, 2, 1, 3).contiguous(), None


class FullAttention(nn.Module):
    def __init__(self, mask_flag=True, factor=5, scale=None, attention_dropout=0.1, output_attention=False):
        super(FullAttention, self).__init__()
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        B, L, H, E = queries.shape
        _, S, _, D = values.shape
        scale = self.scale or 1. / sqrt(E)

        scores = torch.einsum("blhe,bshe->bhls", queries, keys)

        if self.mask_flag:
            if attn_mask is None:
                attn_mask = TriangularCausalMask(B, L, device=queries.device)

            scores.masked_fill_(attn_mask.mask, -np.inf)

        A = self.dropout(torch.softmax(scale * scores, dim=-1))
        V = torch.einsum("bhls,bshd->blhd", A, values)

        if self.output_attention:
            return (V.contiguous(), A)
        else:
            return (V.contiguous(), None)


# Code implementation from https://github.com/zhouhaoyi/Informer2020
class ProbAttention(nn.Module):
    def __init__(self, mask_flag=True, factor=5, scale=None, attention_dropout=0.1, output_attention=False):
        super(ProbAttention, self).__init__()
        self.factor = factor
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)

    def _prob_QK(self, Q, K, sample_k, n_top):  # n_top: c*ln(L_q)
        # Q [B, H, L, D]
        B, H, L_K, E = K.shape
        _, _, L_Q, _ = Q.shape

        # calculate the sampled Q_K
        K_expand = K.unsqueeze(-3).expand(B, H, L_Q, L_K, E)
        # real U = U_part(factor*ln(L_k))*L_q
        index_sample = torch.randint(L_K, (L_Q, sample_k))
        K_sample = K_expand[:, :, torch.arange(
            L_Q).unsqueeze(1), index_sample, :]
        Q_K_sample = torch.matmul(
            Q.unsqueeze(-2), K_sample.transpose(-2, -1)).squeeze()

        # find the Top_k query with sparisty measurement
        M = Q_K_sample.max(-1)[0] - torch.div(Q_K_sample.sum(-1), L_K)
        M_top = M.topk(n_top, sorted=False)[1]

        # use the reduced Q to calculate Q_K
        Q_reduce = Q[torch.arange(B)[:, None, None],
                   torch.arange(H)[None, :, None],
                   M_top, :]  # factor*ln(L_q)
        Q_K = torch.matmul(Q_reduce, K.transpose(-2, -1))  # factor*ln(L_q)*L_k

        return Q_K, M_top

    def _get_initial_context(self, V, L_Q):
        B, H, L_V, D = V.shape
        if not self.mask_flag:
            # V_sum = V.sum(dim=-2)
            V_sum = V.mean(dim=-2)
            contex = V_sum.unsqueeze(-2).expand(B, H,
                                                L_Q, V_sum.shape[-1]).clone()
        else:  # use mask
            # requires that L_Q == L_V, i.e. for self-attention only
            assert (L_Q == L_V)
            contex = V.cumsum(dim=-2)
        return contex

    def _update_context(self, context_in, V, scores, index, L_Q, attn_mask):
        B, H, L_V, D = V.shape

        if self.mask_flag:
            attn_mask = ProbMask(B, H, L_Q, index, scores, device=V.device)
            scores.masked_fill_(attn_mask.mask, -np.inf)

        attn = torch.softmax(scores, dim=-1)  # nn.Softmax(dim=-1)(scores)

        context_in[torch.arange(B)[:, None, None],
        torch.arange(H)[None, :, None],
        index, :] = torch.matmul(attn, V).type_as(context_in)
        if self.output_attention:
            attns = (torch.ones([B, H, L_V, L_V]) /
                     L_V).type_as(attn).to(attn.device)
            attns[torch.arange(B)[:, None, None], torch.arange(H)[
                                                  None, :, None], index, :] = attn
            return (context_in, attns)
        else:
            return (context_in, None)

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        B, L_Q, H, D = queries.shape
        _, L_K, _, _ = keys.shape

        queries = queries.transpose(2, 1)
        keys = keys.transpose(2, 1)
        values = values.transpose(2, 1)

        U_part = self.factor * \
                 np.ceil(np.log(L_K)).astype('int').item()  # c*ln(L_k)
        u = self.factor * \
            np.ceil(np.log(L_Q)).astype('int').item()  # c*ln(L_q)

        U_part = U_part if U_part < L_K else L_K
        u = u if u < L_Q else L_Q

        scores_top, index = self._prob_QK(
            queries, keys, sample_k=U_part, n_top=u)

        # add scale factor
        scale = self.scale or 1. / sqrt(D)
        if scale is not None:
            scores_top = scores_top * scale
        # get the context
        context = self._get_initial_context(values, L_Q)
        # update the context with selected top_k queries
        context, attn = self._update_context(
            context, values, scores_top, index, L_Q, attn_mask)

        return context.contiguous(), attn


class AttentionLayer(nn.Module):
    def __init__(self, attention, d_model, n_heads, d_keys=None,
                 d_values=None):
        super(AttentionLayer, self).__init__()

        d_keys = d_keys or (d_model // n_heads)
        d_values = d_values or (d_model // n_heads)

        self.inner_attention = attention
        self.query_projection = nn.Linear(d_model, d_keys * n_heads)
        self.key_projection = nn.Linear(d_model, d_keys * n_heads)
        self.value_projection = nn.Linear(d_model, d_values * n_heads)
        self.out_projection = nn.Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        H = self.n_heads

        queries = self.query_projection(queries).view(B, L, H, -1)
        keys = self.key_projection(keys).view(B, S, H, -1)
        values = self.value_projection(values).view(B, S, H, -1)

        out, attn = self.inner_attention(
            queries,
            keys,
            values,
            attn_mask,
            tau=tau,
            delta=delta
        )
        out = out.view(B, L, -1)

        return self.out_projection(out), attn

'''
class ReformerLayer(nn.Module):
    def __init__(self, attention, d_model, n_heads, d_keys=None,
                 d_values=None, causal=False, bucket_size=4, n_hashes=4):
        super().__init__()
        self.bucket_size = bucket_size
        self.attn = LSHSelfAttention(
            dim=d_model,
            heads=n_heads,
            bucket_size=bucket_size,
            n_hashes=n_hashes,
            causal=causal
        )

    def fit_length(self, queries):
        # inside reformer: assert N % (bucket_size * 2) == 0
        B, N, C = queries.shape
        if N % (self.bucket_size * 2) == 0:
            return queries
        else:
            # fill the time series
            fill_len = (self.bucket_size * 2) - (N % (self.bucket_size * 2))
            return torch.cat([queries, torch.zeros([B, fill_len, C]).to(queries.device)], dim=1)

    def forward(self, queries, keys, values, attn_mask, tau, delta):
        # in Reformer: defalut queries=keys
        B, N, C = queries.shape
        queries = self.attn(self.fit_length(queries))[:, :N, :]
        return queries, None
'''


import numpy as np
import plotly.express as px
import pandas as pd
from typing import List
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder
from sklearn.preprocessing import * 
from torch.utils.data import DataLoader
from .utils import extend_time_df,MetricsCallback, MyDataset, ActionEnum,beauty_string
from torch.utils.data.sampler import WeightedRandomSampler
try:

    #new version of lightning
    from lightning.pytorch.callbacks import ModelCheckpoint
    import lightning.pytorch as pl
    from ..models.base_v2 import Base
    beauty_string('V2','block',True)
    OLD_PL = False
except:
    ## older version of lightning
    from pytorch_lightning.callbacks import ModelCheckpoint
    import pytorch_lightning as pl
    from ..models.base import Base
    beauty_string('V1','block',True)

    OLD_PL = True

from typing import Union
import os
import torch
import pickle
from datetime import datetime
from ..models.utils import weight_init_zeros,weight_init
import logging 
from .modifiers import *
disable_aim = False

if disable_aim is False:
    from aim.pytorch_lightning import AimLogger
import time
debug_prediction = False
class DummyScaler():
    def __init__(self):
        pass
    def fit(self,x):
        pass
    def transform(self,x):
        return x
    def inverse_transform(self,x):
        return x
    def fit_transform(self,x):
        return x
    

pd.options.mode.chained_assignment = None 
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())      
      
      
class Categorical():
    
    def __init__(self,name:str, frequency: int,duration: List[int], classes: int, action: ActionEnum,  level: List[float]):
        """Class for generating toy categorical data

        Args:
            name (str): name of the categorical signal
            frequency (int): frequency of the signal
            duration (List[int]): duration of each class
            classes (int): number of classes
            action (str): one between additive or multiplicative
            level (List[float]): intensity of each class

        """

        self.name = name 
        self.frequency = frequency
        self.duration = duration
        self.classes = classes
        self.action = action
        self.level = level
        self.validate()

    def validate(self):
        """Validate, maybe there will be other checks in the future
        
        :meta private:
        """
        if len(self.level) == self.classes:
            pass
        else:
            raise ValueError("Length must match")
    
    def generate_signal(self,length:int)->None:
        """Generate the resposne signal

        Args:
            length (int): length of the signal
        """
        if self.action  == 'multiplicative':
            signal = np.ones(length)
        elif self.action == 'additive':
            signal = np.zeros(length)
        classes = []    
        _class = 0
        _level = self.level[0]
        _duration = self.duration[0]

        count = 0
        count_freq = 0
        for i in range(length):
            if count_freq%self.frequency == 0:
                signal[i] = _level
                classes.append(_class)
                count+=1
                if count == _duration:
                    #change class
                    count = 0
                    _class+=1
                    count_freq+= _duration
                    _class = _class%self.classes
                    _level = self.level[_class]
                    _duration = self.duration[_class]

                   
            else:
                classes.append(-1)
                count_freq+=1
        
        self.classes_array = classes
        self.signal_array = signal
        
    def plot(self)->None:
        """Plot the series
        """
        tmp = pd.DataFrame({'time':range(len(self.classes_array)),'signal':self.signal,'class':self.classes_array})
        fig = px.scatter(tmp,x='time',y='signal',color='class',title=self.name)
        fig.show()

 
class TimeSeries():
    
    def __init__(self,name:str,stacked:bool=False):
        """Class for generating time series object. If you don't have any time series you can build one fake timeseries using some helping classes (Categorical for instance).


        Args:
            name (str): name of the series
            stacked (bool): if true it is a stacked model
                
        Usage:
            For example we can generate a toy timeseries:\n
            - add a multiplicative categorical feature (weekly)\n
            >>> settimana = Categorical('settimanale',1,[1,1,1,1,1,1,1],7,'multiplicative',[0.9,0.8,0.7,0.6,0.5,0.99,0.99])\n
            - an additive montly feature (here a year is composed by 5 months)\n
            >>> mese = Categorical('mensile',1,[31,28,20,10,33],5,'additive',[10,20,-10,20,0])\n
            - a spotted categorical variable that happens every 100 days and lasts 1 day\n
            >>> spot = Categorical('spot',100,[7],1,'additive',[10])\n
            >>> ts = TimeSeries('prova')\n
            >>> ts.generate_signal(length = 5000,categorical_variables = [settimana,mese,spot],noise_mean=1,type=0) ##we can add also noise\n
            >>> ts.plot()\n
        """
        self.is_trained = False
        self.name = name
        self.stacked = stacked
        self.verbose = True
        self.group = None
    def __str__(self) -> str:
        return f"Timeseries named {self.name} of length {self.dataset.shape[0]}.\n Categorical variable: {self.cat_var},\n Future variables: {self.future_variables},\n Past variables: {self.past_variables},\n Target variables: {self.target_variables} \n With {'no group' if self.group is None else self.group+' as group' }"
    def __repr__(self) -> str:
        return f"Timeseries named {self.name} of length {self.dataset.shape[0]}.\n Categorical variable: {self.cat_var},\n Future variables: {self.future_variables},\n Past variables: {self.past_variables},\n Target variables: {self.target_variables}\n With {'no group' if self.group is None else self.group+' as group' }"
    
    def set_verbose(self,verbose:bool):
        self.verbose = verbose
    def _generate_base(self,length:int,type:int=0)-> None:
        """Generate a basic timeseries 

        Args:
            length (int): length
            type (int, optional): Type of the generated timeseries. Defaults to 0.
        """
        if type==0:
            self.base_signal = 10*np.cos(np.arange(length)/(2*np.pi*length/100))
            self.out_vars = 1
        else:
            beauty_string('Please implement your own method','block',True)
        """
        
        """    
    def generate_signal(self,length:int=5000,categorical_variables:List[Categorical]=[],noise_mean:int=1,type:int=0)->None:
        """This will generate a syntetic signal with a selected length, a noise level and some categorical variables. The additive series are added at the end while the multiplicative series acts on the original signal
        The TS structure will be populated

        Args:
            length (int, optional): length of the signal. Defaults to 5000.
            categorical_variables (List[Categorical], optional): list of Categorical variables. Defaults to [].
            noise_mean (int, optional): variance of the noise to add at the end. Defaults to 1.
            type (int, optional): type of the timeseries (only type=0 available right now). Defaults to 0.
        """
        
       
        dataset = pd.DataFrame({'time':range(length)})
        self._generate_base(length,type)
        signal = self.base_signal.copy()
        tot = None
        self.cat_var = []
        for c in categorical_variables:
            c.generate_signal(length)
            _signal = c.signal_array
            classes = c.classes_array
            dataset[c.name] = classes
            self.cat_var.append(c.name)
            if c.action=='multiplicative':
                signal*=_signal
            else:
                if tot is None:
                    additive = _signal
                else:
                    additive+=_signal
        signal+=additive
        dataset['signal'] = signal + noise_mean*np.random.randn(len(signal))
        self.dataset = dataset
        
        
        self.past_variables = ['signal']
        self.future_variables = []
        self.target_variables = ['signal']
        self.num_var = list(set(self.past_variables).union(set(self.future_variables)).union(set(self.target_variables)))
        self.num_var = list(np.sort(self.num_var))
        
    def enrich(self,dataset,columns):
        try:
            if columns =='hour':
                dataset[columns] = dataset.time.dt.hour
            elif columns=='dow':
                dataset[columns] = dataset.time.dt.weekday
            elif columns=='month':
                dataset[columns] = dataset.time.dt.month
            elif columns=='minute':
                dataset[columns] = dataset.time.dt.minute
            else:
                if columns  not in dataset.columns:
                    beauty_string(f'I can not automatically enrich column {columns}. Please contact the developers or add it manually to your dataset.','section',True)
        except:
            beauty_string(f'I can not automatically enrich column {columns}. Probably not a temporal index.','section',True)

    def load_signal(self,data:pd.DataFrame,
                    enrich_cat:List[str] = [],
                    past_variables:List[str]=[],
                    future_variables:List[str]=[],
                    target_variables:List[str]=[],
                    cat_past_var:List[str]=[],
                    cat_fut_var:List[str]=[],
                    check_past:bool=True,
                    group:Union[None,str]=None,
                    check_holes_and_duplicates:bool=True,
                    silly_model:bool=False,
                    sampler_weights:Union[None,str]=None)->None:
        """ This is a crucial point in the data structure. We expect here to have a dataset with time as timestamp.
            There are some checks:
                1- the duplicates will tbe removed taking the first instance
                
                2- the frequency will the inferred taking the minumum time distance between samples
               
                3- the dataset will be filled completing the missing timestamps

        Args:
            data (pd.DataFrame): input dataset the column indicating the time must be called `time`
            enrich_cat (List[str], optional): it is possible to let this function enrich the dataset for example adding the standard columns: hour, dow, month and minute. Defaults to [].
            past_variables (List[str], optional): list of column names of past variables not available for future times . Defaults to [].
            future_variables (List[str], optional): list of future variables available for tuture times. Defaults to [].
            target_variables (List[str], optional): list of the target variables. They will added to past_variables by default unless `check_past` is false. Defaults to [].
            cat_past_var (List[str], optional): list of the past categorical variables. Defaults to [].
            cat_future_var (List[str], optional): list of the future categorical variables. Defaults to [].
            check_past (bool, optional): see `target_variables`. Defaults to True.
            group (str or None, optional): if not None the time serie dataset is considered composed by omogeneus timeseries coming from different realization (for example point of sales, cities, locations) and the relative series are not splitted during the sample generation. Defaults to None
            check_holes_and_duplicates (bool, optional): if False duplicates or holes will not checked, the dataloader can not correctly work, disable at your own risk. Defaults True
            silly_model (bool, optional): if True, target variables will be added to the pool of the future variables. This can be useful to see if information passes throught the decoder part of your model (if any)
            sampler_weights  group (str or None, optional): if it is a column name it will be used as weight for the sampler. Careful that the weight of the sample is the weight value of the fist target value (index)
        """
        
        
        
        dataset = data.copy()
        dataset.sort_values(by='time',inplace=True)
        
        if check_holes_and_duplicates:
            beauty_string('I will drop duplicates, I dont like them','section',self.verbose)
            dataset.drop_duplicates(subset=['time'] if group is None else [group,'time'],  keep='first', inplace=True, ignore_index=True)
            
            if group is None:
                differences = dataset.time.diff()[1:]
            else:
                differences = dataset[dataset[group]==dataset[group].unique()[0]].time.diff()[1:]
                
     
            if isinstance(dataset.time[0], datetime):
                freq = pd.to_timedelta(differences.min())   
            else:
                if int(dataset.time[0])==dataset.time[0]: ##ONLY THINK THAT WORKS IN GENERAL
                    freq = int(differences.min())
                else:
                    raise TypeError("time must be integer or datetime")
            self.freq = freq 
            
            
            if differences.nunique()>1:
                beauty_string("There are holes in the dataset i will try to extend the dataframe inserting NAN",'info',self.verbose)
                beauty_string(f'Detected minumum frequency: {freq}','section',self.verbose)
                dataset = extend_time_df(dataset,freq,group).merge(dataset,how='left')
        else:
            beauty_string("I will compute the frequency as minimum of the time difference",'info',self.verbose)
            self.freq =  dataset.time.diff()[1:].min()
            if isinstance(dataset.time.dtype, datetime):
                self.freq = pd.to_timedelta(self.freq)   
            
                
        assert len(target_variables)>0, 'Provide at least one column for target'
        assert 'time'  in dataset.columns, 'The temporal column must be called time'
        if set(target_variables).intersection(set(past_variables))!= set(target_variables): 
            if check_past:
                beauty_string('I will update past column adding all target columns, if you want to avoid this beahviour please use check_pass as false','info',self.verbose)
                past_variables = list(set(past_variables).union(set(target_variables)))
                past_variables = list(np.sort(past_variables))
        self.cat_past_var = cat_past_var
        self.cat_fut_var = cat_fut_var

        self.group = group 
        if group is not None:
            if group not in cat_past_var:
                beauty_string(f'I will add {group} to the categorical past/future variables','info',self.verbose)
                self.cat_past_var.append(group)
            if group not in cat_fut_var:
                beauty_string(f'I will add {group} to the categorical past/future variables','info',self.verbose)
                self.cat_fut_var.append(group)   
                
        self.enrich_cat = enrich_cat
        for c in enrich_cat:
            self.cat_past_var = list(set(self.cat_past_var+[c]))
            self.cat_fut_var = list(set(self.cat_fut_var+[c]))  
            if c in dataset.columns:
                beauty_string('Categorical {c} already present, it will be added to categorical variable but not call the enriching function','info',self.verbose) 
            else:
                self.enrich(dataset,c)
        self.cat_past_var = list(np.sort(self.cat_past_var))
        self.cat_fut_var = list(np.sort(self.cat_fut_var))
        
        self.cat_var = list(set(self.cat_past_var+self.cat_fut_var)) ## all categorical data
        self.cat_var = list(np.sort(self.cat_var))
        self.dataset = dataset
        self.past_variables = past_variables
        self.future_variables = future_variables
        self.target_variables = target_variables
        self.out_vars = len(target_variables)
        self.num_var = list(set(self.past_variables).union(set(self.future_variables)).union(set(self.target_variables)))
        self.num_var = list(np.sort(self.num_var))
        if silly_model:
            beauty_string('YOU ARE TRAINING A SILLY MODEL WITH THE TARGETS IN THE INPUTS','section',self.verbose) 
            self.future_variables+=self.target_variables
        self.sampler_weights = sampler_weights
    def plot(self):
        """  
        Easy way to control the loaded data
        Returns:
            plotly.graph_objects._figure.Figure: figure of the target variables
        """
      
        beauty_string('Plotting only target variables','block',self.verbose)
        if self.group is None:
            tmp = self.dataset[['time']+self.target_variables].melt(id_vars=['time'])
            fig = px.line(tmp,x='time',y='value',color='variable',title=self.name)
            fig.show()
        else:
            tmp = self.dataset[['time',self.group]+self.target_variables].melt(id_vars=['time',self.group])
            fig = px.line(tmp,x='time',y='value',color='variable',title=self.name,facet_row=self.group)
            fig.show()
        return fig
    
    logging.getLogger("lightning").setLevel(logging.CRITICAL)
    logging.getLogger("lightning.pytorch").setLevel(logging.CRITICAL)
    logging.getLogger("lightning.pytorch.utilities.rank_zero").setLevel(logging.CRITICAL)
    logging.getLogger("lightning.pytorch.trainer").setLevel(logging.CRITICAL)
    
    def create_data_loader(self,data:pd.DataFrame,
                           past_steps:int,
                           future_steps:int,
                           shift:int=0,
                           keep_entire_seq_while_shifting:bool=False,
                           starting_point:Union[None,dict]=None,
                           skip_step:int=1,
                           is_inference:bool=False
                         
                           )->MyDataset:
        """ Create the dataset for the training/inference step

        Args:
            data (pd.DataFrame): input dataset, usually a subset of self.data
            past_steps (int): past context length
            future_steps (int): future lags to predict
            shift (int, optional): if >0 the future input variables will be shifted (categorical and numerical). For example for attention model it is better to start with a know value of y and use it during the process. Defaults to 0.
            keep_entire_seq_while_shifting (bool, optional): if the dataset is shifted, you may want the future data be of length future_step+shift (like informer), default false
            starting_point (Union[None,dict], optional): a dictionary indicating if a sample must be considered. It is checked for the first lag in the future (useful in the case your model has to predict only starting from hour 12). Defaults to None.
            skip_step (int, optional): list of the categortial variables (same for past and future). Usual there is a skip of one between two saples but for debugging  or training time purposes you can skip some samples. Defaults to 1.
        Returns:
            MyDataset: class that extends torch.utils.data.Dataset (see utils)
                keys of a batch:
                y : the target variable(s)
                x_num_past: the numerical past variables
                x_num_future: the numerical future variables
                x_cat_past: the categorical past variables
                x_cat_future: the categorical future variables
                idx_target: index of target features in the past array
        """
        beauty_string('Creating data loader','block',self.verbose)
        
        x_num_past_samples = []
        x_num_future_samples = []
        x_cat_past_samples = []
        x_cat_future_samples = []
        y_samples = []
        t_samples = []
        g_samples = []
        sampler_weights_samples = []
        
        if starting_point is not None:
            kk = list(starting_point.keys())[0]
            assert kk not in self.cat_var, beauty_string('CAN NOT USE FEATURE {kk} as starting point it may have a different value due to the normalization step, please add a second column with a suitable name','info',True)
        
        ##overwrite categorical columns
        for c in self.cat_var:
            self.enrich(data,c)

        if self.group is None:
            data['_GROUP_'] = '1'
        else:
            data['_GROUP_'] = data[self.group].values
            
            
        if self.normalize_per_group:
            tot = []
            groups = data[self.group].unique()
            
            data[self.group] = self.scaler_cat[self.group].transform(data[self.group].values.reshape(-1,1)).flatten()
            
            for group in groups:
                tmp = data[data['_GROUP_']==group].copy()
                
                for c in self.num_var:
                    tmp[c] = self.scaler_num[f'{c}_{group}'].transform(tmp[c].values.reshape(-1,1)).flatten()
                for c in self.cat_var:      
                    if c!=self.group:                         
                        tmp[c] = self.scaler_cat[f'{c}_{group}'].transform(tmp[c].values.reshape(-1,1)).flatten()
                tot.append(tmp)
            data = pd.concat(tot,ignore_index=True)
        else:
            for c in self.cat_var:
                data[c] = self.scaler_cat[c].transform(data[c].values.reshape(-1,1)).flatten()
            for c in self.num_var: 
                data[c] = self.scaler_num[c].transform(data[c].values.reshape(-1,1)).flatten()

        idx_target = []
        for c in self.target_variables:
            if c in self.past_variables:
                idx_target.append(self.past_variables.index(c))
            
        idx_target_future = []
        
        for c in self.target_variables:
            if c in self.future_variables:
                idx_target_future.append(self.future_variables.index(c))    
        if len(idx_target_future)==0:
            idx_target_future = None
        

        if self.stacked:
            skip_stacked = future_steps*future_steps-future_steps
        else:
            skip_stacked = 0
        for group in data['_GROUP_'].unique():
            tmp = data[data['_GROUP_']==group]
            groups = tmp['_GROUP_'].values  
            t = tmp.time.values 
            x_num_past = tmp[self.past_variables].values
            if len(self.future_variables)>0:
                x_num_future = tmp[self.future_variables].values
            if len(self.cat_past_var)>0:
                x_past_cat = tmp[self.cat_past_var].values
            if len(self.cat_fut_var)>0:
                x_fut_cat = tmp[self.cat_fut_var].values
            y_target = tmp[self.target_variables].values
            if self.sampler_weights is not None:
                sampler_weights = tmp[self.sampler_weights].values.flatten()
        
            if starting_point is not None:
                check = tmp[list(starting_point.keys())[0]].values == starting_point[list(starting_point.keys())[0]]
            else:
                check = [True]*len(y_target)
            
            for i in range(past_steps,tmp.shape[0]-future_steps-skip_stacked,skip_step):
                if check[i]:

                    if len(self.future_variables)>0:
                        if keep_entire_seq_while_shifting:
                            xx = x_num_future[i-shift+skip_stacked:i+future_steps+skip_stacked].mean()
                        else:
                            xx = x_num_future[i-shift+skip_stacked:i+future_steps-shift+skip_stacked].mean()
                    else:
                        xx = 0.0
                    if is_inference is False:
                        xx+=y_target[i+skip_stacked:i+future_steps+skip_stacked].min()
                    
                    if np.isfinite(x_num_past[i-past_steps:i].min()  + xx):
                        
                        x_num_past_samples.append(x_num_past[i-past_steps:i])
                        if len(self.future_variables)>0:
                            if keep_entire_seq_while_shifting:
                                x_num_future_samples.append(x_num_future[i-shift+skip_stacked:i+future_steps+skip_stacked])
                            else:
                                x_num_future_samples.append(x_num_future[i-shift+skip_stacked:i+future_steps-shift+skip_stacked])
                        if len(self.cat_past_var)>0:
                            x_cat_past_samples.append(x_past_cat[i-past_steps:i])
                        if len(self.cat_fut_var)>0:
                            if keep_entire_seq_while_shifting:
                                x_cat_future_samples.append(x_fut_cat[i-shift+skip_stacked:i+future_steps+skip_stacked])
                            else:
                                x_cat_future_samples.append(x_fut_cat[i-shift+skip_stacked:i+future_steps-shift+skip_stacked])
                        
                        y_samples.append(y_target[i+skip_stacked:i+future_steps+skip_stacked])
                        if self.sampler_weights is not None:
                            sampler_weights_samples.append(sampler_weights[i+skip_stacked])
                        t_samples.append(t[i+skip_stacked:i+future_steps+skip_stacked])
                        g_samples.append(groups[i])

    
        
        if len(self.future_variables)>0:
            try:
                x_num_future_samples = np.stack(x_num_future_samples)
            except Exception as e:
                beauty_string('WARNING x_num_future_samples is empty and it should not','info',True)
        
        y_samples = np.stack(y_samples)
        if self.sampler_weights is not None:
            sampler_weights_samples = np.stack(sampler_weights_samples)
        t_samples = np.stack(t_samples)   
        g_samples = np.stack(g_samples)

        if len(self.cat_past_var)>0:
            x_cat_past_samples = np.stack(x_cat_past_samples).astype(np.int32)
        if len(self.cat_fut_var)>0:
            x_cat_future_samples = np.stack(x_cat_future_samples).astype(np.int32)
        x_num_past_samples = np.stack(x_num_past_samples)
        if self.stacked:
            mod = 0
        else:
            mod = 1.0
        dd = {'y':y_samples.astype(np.float32),
              'x_num_past':(x_num_past_samples*mod).astype(np.float32)}
        if len(self.cat_past_var)>0:
            dd['x_cat_past'] = x_cat_past_samples
        if len(self.cat_fut_var)>0:
            dd['x_cat_future'] = x_cat_future_samples
        if len(self.future_variables)>0:
            dd['x_num_future'] = x_num_future_samples.astype(np.float32)
        if self.sampler_weights is not None:
            dd['sampler_weights'] = sampler_weights_samples.astype(np.float32)
        else:
            dd['sampler_weights'] = np.ones(len(y_samples)).astype(np.float32)
        return MyDataset(dd,t_samples,g_samples,idx_target,idx_target_future)
    
          
    
    def split_for_train(self,
                        perc_train:Union[float,None]=0.6,
                        perc_valid:Union[float,None]=0.2,
                        range_train:Union[List[Union[datetime, str]],None]=None,
                        range_validation:Union[List[Union[datetime, str]],None]=None,
                        range_test:Union[List[Union[datetime, str]],None]=None,
                        past_steps:int = 100,
                        future_steps:int=20,
                        shift:int = 0,
                        keep_entire_seq_while_shifting:bool=False,
                        starting_point:Union[None, dict]=None,
                        skip_step:int=1,
                        normalize_per_group: bool=False,
                        check_consecutive: bool=True,
                        scaler: str='StandardScaler()' 
                        )->List[DataLoader]:
        """Split the data and create the datasets.

        Args:
            perc_train (Union[float,None], optional): fraction of the training set. Defaults to 0.6.
            perc_valid (Union[float,None], optional): fraction of the test set. Defaults to 0.2.
            range_train (Union[List[Union[datetime, str]],None], optional): a list of two elements indicating the starting point and end point of the training set (string date style or datetime). Defaults to None.
            range_validation (Union[List[Union[datetime, str]],None], optional):a list of two elements indicating the starting point and end point of the validation set (string date style or datetime). Defaults to None.
            range_test (Union[List[Union[datetime, str]],None], optional): a list of two elements indicating the starting point and end point of the test set (string date style or datetime). Defaults to None.
            past_steps (int, optional): past step to consider for making the prediction. Defaults to 100.
            future_steps (int, optional): future step to predict. Defaults to 20.
            shift (int, optional): see `create_data_loader`. Defaults to 0.
            keep_entire_seq_while_shifting (bool, optional): if the dataset is shifted, you may want the future data be of length future_step+shift (like informer), default false

            starting_point (Union[None, dict], optional):  see `create_data_loader`. Defaults to None.
            skip_step (int, optional):  see `create_data_loader`. Defaults to 1.
            normalize_per_group (boolean, optional): if true and self.group is not None, the variables are scaled respect to the groups. Default False
            check_consecutive (boolean, optional): if false it skips the check on the consecutive ranges. Default True
            scaler: instance of a sklearn.preprocessing scaler. Default 'StandardScaler()'
        Returns:
            List[DataLoader,DataLoader,DataLoadtrainer]: three dataloader used for training or inference
        """

        beauty_string('Splitting for train','block',self.verbose)

        
        try:
            ls = self.dataset.shape[0]
        except Exception as _:
            beauty_string('Empty dataset','info', True)
            return None, None, None
        
        if range_train is None:
            if self.group is None:
                beauty_string(f'Split temporally using perc_train: {perc_train} and perc_valid:{perc_valid}','section',self.verbose)
                train = self.dataset.iloc[0:int(perc_train*ls)]
                validation = self.dataset.iloc[int(perc_train*ls):int(perc_train*ls+perc_valid*ls)]
                test = self.dataset.iloc[int(perc_train*ls+perc_valid*ls):]
            else:
                beauty_string(f'Split temporally using perc_train: {perc_train} and perc_valid:{perc_valid} for each group!','info',self.verbose)
                train = []
                validation =[]
                test = []
                ls = self.dataset.groupby(self.group).time.count().reset_index()
                for group in self.dataset[self.group].unique():
                    tmp = self.dataset[self.dataset[self.group]==group]
                    lt = ls[ls[self.group]==group].time.values[0]
                    train.append(tmp[0:int(perc_train*lt)])
                    validation.append(tmp[int(perc_train*lt):int(perc_train*lt+perc_valid*lt)])
                    test.append(tmp[int(perc_train*lt+perc_valid*lt):])

                train = pd.concat(train,ignore_index=True)
                validation = pd.concat(validation,ignore_index=True)
                test = pd.concat(test,ignore_index=True)


        else:
            if check_consecutive:
                assert range_train[0]<range_train[1]<=range_validation[0]<range_validation[1]<=range_test[0]<range_test[1], beauty_string(f'The range are not correct','info',True)
            beauty_string('Split temporally using the time intervals provided','section',self.verbose)
            train = self.dataset[self.dataset.time.between(range_train[0],range_train[1])]
            validation =  self.dataset[self.dataset.time.between(range_validation[0],range_validation[1])]
            test =  self.dataset[self.dataset.time.between(range_test[0],range_test[1])]


        beauty_string('Train categorical and numerical scalers','block',self.verbose)

        if self.is_trained:
            pass
        else:
            self.scaler_cat = {}
            self.scaler_num = {}
            if self.group is None or normalize_per_group is False:
                self.normalize_per_group = False
                for c in self.num_var:
                    self.scaler_num[c] =  eval(scaler)
                    self.scaler_num[c].fit(train[c].values.reshape(-1,1))
                for c in self.cat_var:                               
                    self.scaler_cat[c] =  OrdinalEncoder(dtype=np.int32,handle_unknown= 'use_encoded_value',unknown_value=train[c].nunique())
                   
                    self.scaler_cat[c].fit(train[c].values.reshape(-1,1))  
            else:
                self.normalize_per_group = True
                self.scaler_cat[self.group] =  OrdinalEncoder(dtype=np.int32,handle_unknown= 'use_encoded_value',unknown_value=train[c].nunique())
                self.scaler_cat[self.group].fit(train[self.group].values.reshape(-1,1))  
                for group in train[self.group].unique():
                    tmp = train[train[self.group]==group]

                    for c in self.num_var:
                        self.scaler_num[f'{c}_{group}'] =  eval(scaler)
                        self.scaler_num[f'{c}_{group}'].fit(tmp[c].values.reshape(-1,1))
                    for c in self.cat_var:
                        if c!=self.group:                               
                            self.scaler_cat[f'{c}_{group}'] =  OrdinalEncoder(dtype=np.int32,handle_unknown= 'use_encoded_value',unknown_value=train[c].nunique())
                            self.scaler_cat[f'{c}_{group}'].fit(tmp[c].values.reshape(-1,1))  
 
        dl_train = self.create_data_loader(train,past_steps,future_steps,shift,keep_entire_seq_while_shifting,starting_point,skip_step)
        dl_validation = self.create_data_loader(validation,past_steps,future_steps,shift,keep_entire_seq_while_shifting,starting_point,skip_step)
        if test.shape[0]>0:
            dl_test = self.create_data_loader(test,past_steps,future_steps,shift,keep_entire_seq_while_shifting,starting_point,skip_step)
        else:
            dl_test = None
        return dl_train,dl_validation,dl_test
            
    def set_model(self,model:Base,config:dict=None,custom_init:bool=False):
        """Set the model to train

        Args:
            model (Base): see `models`
            config (dict, optional): usually the configuration used by the model. Defaults to None.
            custom_init (bool, optional): if true a custom initialization paradigm will be used (see weight_init in models/utils.py ) .
        """
        self.model = model
        if custom_init:
            self.model.apply(weight_init)
        #self.model.apply(weight_init_zeros)

        self.config = config


        beauty_string('Setting the model','block',self.verbose)
        beauty_string(model,'',self.verbose)
        
    def train_model(self,dirpath:str,
                    split_params:dict,
                    batch_size:int=100,
                    num_workers:int=4,
                    max_epochs:int=500,
                    auto_lr_find:bool=True,
                    gradient_clip_val:Union[float,None]=None,
                    gradient_clip_algorithm:str="value",
                    devices:Union[str,List[int]]='auto',
                    precision:Union[str,int]=32,
                    modifier:Union[None,str]=None,
                    modifier_params:Union[None,dict]=None,
                    seed:int=42
                    )-> float:
        """Train the model

        Args:
            dirpath (str): path where to put all the useful things
            split_params (dict): see `split_for_train`
            batch_size (int, optional): batch size. Defaults to 100.
            num_workers (int, optional): num_workers for the dataloader. Defaults to 4.
            max_epochs (int, optional): maximum epochs to perform. Defaults to 500.
            auto_lr_find (bool, optional): find initial learning rate, see  `pytorch-lightening`. Defaults to True.
            gradient_clip_val (Union[float,None], optional): gradient_clip_val. Defaults to None. See https://lightning.ai/docs/pytorch/stable/advanced/training_tricks.html
            gradient_clip_algorithm (str, optional): gradient_clip_algorithm. Defaults to 'norm '. See https://lightning.ai/docs/pytorch/stable/advanced/training_tricks.html
            devices (Union[str,List[int]], optional): devices to use. Use auto if cpu or the list of gpu to use otherwise. Defaults to 'auto'.
            precision  (Union[str,int], optional): precision to use. Usually 32 bit is fine but for larger model you should try 'bf16'. If 'auto' it will use bf16 for GPU and 32 for cpu
            modifier (Union[str,int], optional): if not None a modifier is applyed to the dataloader. Sometimes lightening has very restrictive rules on the dataloader, or we want to use a ML model before or after the DL model (See readme for more information)
            modifier_params (Union[dict,int], optional): parameters of the modifier
            seed (int, optional): seed for reproducibility
        """

        beauty_string('Training the model','block',self.verbose)

        self.split_params = split_params
        self.check_custom = False
        train,validation,test = self.split_for_train(**self.split_params)
        accelerator = 'gpu' if torch.cuda.is_available() else "cpu"
        strategy = "auto"
        if accelerator == 'gpu':
            strategy = "auto" ##TODO in future investigate on this
            if precision=='auto':
                precision = 'bf16'
            #"bf16" ##in futuro magari inserirlo nei config, potrebbe essere che per alcuni modelli possa non andare bfloat32
            torch.set_float32_matmul_precision('medium')
            beauty_string('Setting multiplication precision to medium','info',self.verbose)
        else:
            devices = 'auto'
            if precision=='auto':
                precision  = 32
        beauty_string(f'train:{len(train)}, validation:{len(validation)}, test:{len(test) if test is not None else 0}','section',self.verbose)
        if (accelerator=='gpu') and (num_workers>0):
            persistent_workers = True
        else:
            persistent_workers = False
            
            
        if modifier is not None:
            modifier = eval(modifier)
            modifier = modifier(**modifier_params)
            train, validation = modifier.fit_transform(train=train,val=validation)
            self.modifier = modifier
        else:
            self.modifier = None
        
        if self.sampler_weights is not None:
            beauty_string(f'USING SAMPLER IN TRAIN {min(train.sampler_weights)}-{max(train.sampler_weights)}','section',self.verbose)

            sampler = WeightedRandomSampler(train.sampler_weights, num_samples= len(train))
            train_dl = DataLoader(train, batch_size = batch_size , shuffle=False,sampler=sampler,drop_last=True,num_workers=num_workers,persistent_workers=persistent_workers)

        else:
            train_dl = DataLoader(train, batch_size = batch_size , shuffle=True,drop_last=True,num_workers=num_workers,persistent_workers=persistent_workers)
        valid_dl = DataLoader(validation, batch_size = batch_size , shuffle=False,drop_last=False,num_workers=num_workers,persistent_workers=persistent_workers)
        
        beauty_string(f'train:{len(train_dl)}, validation:{len(valid_dl)}','section',self.verbose)

        if debug_prediction:
            dl = DataLoader(test, batch_size = batch_size , shuffle=False,drop_last=True,num_workers=num_workers,persistent_workers=persistent_workers)
            res = []
            real = []
            for batch in dl:
        

                res.append(self.model.inference(batch).cpu().detach().numpy())
                real.append(batch['y'].cpu().detach().numpy())

            res = np.vstack(res)
            real = np.vstack(real)
            with open('/home/agobbi/Projects/ExpTS/tmp_beginning.pkl','wb') as f:
                import pickle
                pickle.dump([res,real],f)
        
        checkpoint_callback = ModelCheckpoint(dirpath=dirpath,
                                     monitor='val_loss',
                                      save_last = True,
                                      every_n_epochs =1,
                                      verbose = self.verbose,
                                      save_top_k = 1,
                                     filename='checkpoint')
        
        
        #logger = CSVLogger("logs", name=dirpath)
        beauty_string(f'Init aim','section',self.verbose)


        #https://stackoverflow.com/questions/49201236/check-the-total-number-of-parameters-in-a-pytorch-model
        n_params = sum(dict((p.data_ptr(), p.numel()) for p in self.model.parameters()).values())

        #https://discuss.pytorch.org/t/finding-model-size/130275
        param_size = 0
        for param in self.model.parameters():
            param_size += param.nelement() * param.element_size()
        buffer_size = 0
        for buffer in self.model.buffers():
            buffer_size += buffer.nelement() * buffer.element_size()

        size_all_mb = (param_size + buffer_size) / 1024**2
        tmp = self.config.copy()
        tmp['model_name'] = self.model.name

        #aim_logger.experiment.track(self.model.name,name='model_name')
        if disable_aim is False:
            aim_logger = AimLogger(
                experiment=self.name,
                train_metric_prefix='train_',
                val_metric_prefix='val_',
            )
            aim_logger.experiment.track(n_params,name='N-parameters')
            aim_logger.experiment.track(size_all_mb,name='dim-model-MB')
            aim_logger.experiment.track(len(train_dl.dataset),name='len-train')
            aim_logger.experiment.track(len(valid_dl.dataset),name='len-valid')
            #aim_logger.experiment.track(self.config,name=None)
            
            aim_logger._run['hyperparameters'] = tmp
        else:
            aim_logger = False
        mc = MetricsCallback(dirpath)
        ## TODO se ci sono 2 o piu gpu MetricsCallback non funziona (secondo me fa una istanza per ogni dataparallel che lancia e poi non riesce a recuperare info)
        pl.seed_everything(seed, workers=True)
        self.model.max_epochs = max_epochs
        
        if os.path.isfile(os.path.join(dirpath,'last.ckpt')):
            weight_exists = True
            beauty_string('I loaded a previous checkpoint','section',self.verbose)

        else: 
            weight_exists = False
            beauty_string('I can not load a previous model','section',self.verbose)

        self.model.to(torch.device("cuda:0" if torch.cuda.is_available() else "cpu"))
        if self.model.can_be_compiled():
            try:
                self.model = torch.compile(self.model)
                beauty_string('Model COMPILED','block',self.verbose)

            except:
                beauty_string('Can not compile the model','block',self.verbose)
        else:
            beauty_string('Model can not still be compiled, be patient','block',self.verbose)

        
        if OLD_PL:
            trainer = pl.Trainer(default_root_dir=dirpath,
                                logger = aim_logger,
                                max_epochs=max_epochs,
                                callbacks=[checkpoint_callback,mc],
                                auto_lr_find=auto_lr_find, 
                                accelerator=accelerator,
                                log_every_n_steps=1,
                                devices=devices,
                                strategy=strategy,
                                enable_progress_bar=False,
                                precision=precision,
                                gradient_clip_val=gradient_clip_val,
                                gradient_clip_algorithm=gradient_clip_algorithm)#,devices=1)
        else:
            trainer = pl.Trainer(default_root_dir=dirpath,
                                logger = aim_logger,
                                max_epochs=max_epochs,
                                callbacks=[checkpoint_callback,mc],
                                strategy='auto',
                                devices=devices,
                                log_every_n_steps=5,
                                enable_progress_bar=False,
                                precision=precision,
                                gradient_clip_val=gradient_clip_val,
                                gradient_clip_algorithm=gradient_clip_algorithm)#,devices=1)
        tot_seconds = time.time()


        beauty_string(f'tuning lr','section',self.verbose)

     
        if auto_lr_find and (weight_exists is False):
            try:
                if OLD_PL:
                    lr_tuner = trainer.tune(self.model,train_dataloaders=train_dl,val_dataloaders = valid_dl)
                    files = os.listdir(dirpath)
                    for f in files:
                        if '.lr_find' in f:
                            os.remove(os.path.join(dirpath,f))
                    self.model.optim_config['lr'] = lr_tuner['lr_find'].suggestion()
                else:
                    from lightning.pytorch.tuner import Tuner
                    tuner = Tuner(trainer)
                    lr_finder = tuner.lr_find(self.model,train_dataloaders=train_dl,val_dataloaders = valid_dl)
                    self.model.optim_config['lr'] = lr_finder.suggestion() ## we are using it as optim key
            except Exception as e:
                beauty_string(f'There is a problem with the finding LR routine {e}','section',self.verbose)
        beauty_string(f'Start fitting','section',self.verbose)

        if OLD_PL:
            if weight_exists:
                trainer.fit(self.model, train_dl,valid_dl,ckpt_path=os.path.join(dirpath,'last.ckpt'))
            else:
                trainer.fit(self.model, train_dl,valid_dl)
        else:
            if weight_exists:
                trainer.fit(self.model, train_dataloaders = train_dl,val_dataloaders = valid_dl,ckpt_path=os.path.join(dirpath,'last.ckpt'))
            else:
                trainer.fit(self.model, train_dataloaders = train_dl,val_dataloaders = valid_dl)

        self.checkpoint_file_best = checkpoint_callback.best_model_path
        self.checkpoint_file_last = checkpoint_callback.last_model_path 
        if self.checkpoint_file_last=='':
            beauty_string('There is a bug on saving last model I will try to fix it','info',self.verbose)
            self.checkpoint_file_last = os.path.join(dirpath, "last.ckpt")
            trainer.save_checkpoint(os.path.join(dirpath, "last.ckpt"))
        if self.checkpoint_file_best=='':
            beauty_string('There is a bug on saving best model I will try to fix it','info',self.verbose)
            self.checkpoint_file_best = os.path.join(dirpath, "checkpoint.ckpt")
            trainer.save_checkpoint(os.path.join(dirpath, "checkpoint.ckpt"))

        self.dirpath = dirpath
        
        self.losses = mc.metrics

        files = os.listdir(dirpath)
        if debug_prediction:
            res = []
            real = []
            for batch in dl:
                res.append(self.model.inference(batch).cpu().detach().numpy())
                real.append(batch['y'].cpu().detach().numpy())

            res = np.vstack(res)
            real = np.vstack(real)
            with open('/home/agobbi/Projects/ExpTS/tmp_after_training.pkl','wb') as f:
                import pickle
                pickle.dump([res,real],f)
        
        ##accrocchio per multi gpu
        for f in files:
            if '__losses__.csv' in f:
                if len(self.losses['val_loss'])>0:
                    self.losses = pd.DataFrame(self.losses)
                else:
                    self.losses = pd.read_csv(os.path.join(os.path.join(dirpath,f)))
                os.remove(os.path.join(os.path.join(dirpath,f)))
        if isinstance(self.losses,dict):
            self.losses = pd.DataFrame()

        try:
            if OLD_PL:
                if isinstance(self.model, torch._dynamo.eval_frame.OptimizedModule):
                    self.model = self.model._orig_mod 
                    self.model.load_from_checkpoint(self.checkpoint_file_last)
                else:
                    self.model = self.model.load_from_checkpoint(self.checkpoint_file_last)
            else:
                if isinstance(self.model, torch._dynamo.eval_frame.OptimizedModule):
                    mm = self.model._orig_mod 
                    self.model = mm.__class__.load_from_checkpoint(self.checkpoint_file_last)
                else:
                    self.model = self.model.__class__.load_from_checkpoint(self.checkpoint_file_last)
            if debug_prediction:
                res = []
                real = []
                for batch in dl:
            

                    res.append(self.model.inference(batch).cpu().detach().numpy())
                    real.append(batch['y'].cpu().detach().numpy())

                res = np.vstack(res)
                real = np.vstack(real)
                with open('/home/agobbi/Projects/ExpTS/tmp_after_loading.pkl','wb') as f:
                    import pickle
                    pickle.dump([res,real],f)
            
        except Exception as _:
            beauty_string(f'There is a problem loading the weights on file MAYBE CHANGED HOW WEIGHTS ARE LOADED {self.checkpoint_file_last}','section',self.verbose)

        try:
            val_loss = self.losses.val_loss.values[-1]
        except Exception as _:
            beauty_string('Can not extract the validation loss, maybe it is a persistent model','info',self.verbose)
            val_loss = 100
        self.is_trained = True
        
        beauty_string('END of the training process','block',self.verbose)


        if disable_aim is False:
            aim_logger.experiment.track((time.time()-tot_seconds),name='seconds-training')
            aim_logger.experiment.track(val_loss,name='val-loss-end-train')

        
        
        return val_loss 
    
    def inference_on_set(self,batch_size:int=100,
                         num_workers:int=4,
                         split_params:Union[None,dict]=None,set:str='test',
                         rescaling:bool=True,
                         data:Union[None,torch.utils.data.Dataset]=None)->pd.DataFrame:
        """This function allows to get the prediction on a particular set (train, test or validation). 

        Args:
            batch_size (int, optional): barch sise. Defaults to 100.
            num_workers (int, optional): num workers. Defaults to 4.
            split_params (Union[None,dict], optional): if not None  the spliting procedure will use the given data otherwise it will use the same configuration used in train. Defaults to None.
            set (str, optional): trai, validation or test. Defaults to 'test'.
            rescaling (bool, optional):  If rescaling is true the output will be rescaled to the initial values. . Defaults to True.
            data (None or pd.DataFrame, optional). If not None the inference is performed on the given data. In the case of custom data please call inference because it will normalize the data for you!
        Returns:
            pd.DataFrame: the predicted values in a pandas format
        """
        
        beauty_string('Inference on a set (train, validation o test)','block',self.verbose)
     
        if data is None:
            if split_params is None:
                beauty_string(f'splitting using train parameters {self.split_params}','section',self.verbose)
                train,validation,test = self.split_for_train(**self.split_params)
            else:
                train,validation,test = self.split_for_train(**split_params)

        if set=='test':
            if self.modifier is not None:
                test = self.modifier.transform(test)
            dl = DataLoader(test, batch_size = batch_size , shuffle=False,drop_last=False,num_workers=num_workers)
        elif set=='validation':
            if self.modifier is not None:
                validation = self.modifier.transform(validation)
            dl = DataLoader(validation, batch_size = batch_size , shuffle=False,drop_last=False,num_workers=num_workers)
        elif set=='train':
            if self.modifier is not None:
                train = self.modifier.transform(train)
            dl = DataLoader(train, batch_size = batch_size , shuffle=False,drop_last=False,num_workers=num_workers)    
        elif set=='custom':
            if self.check_custom:
                pass
            else:
                beauty_string('If you are here something went wrong, please report it','section',self.verbose)
            if self.modifier is not None:
                data = self.modifier.transform(data)
            dl = DataLoader(data, batch_size = batch_size , shuffle=False,drop_last=False,num_workers=num_workers)    
  
        else:
            beauty_string('Select one of train, test, or validation set','section',self.verbose)
        self.model.eval()
        
        res = []
        real = []
        self.model.to(torch.device("cuda:0" if torch.cuda.is_available() else "cpu"))
        beauty_string(f'Device used: {self.model.device}','info',self.verbose)

        for batch in dl:
    

            res.append(self.model.inference(batch).cpu().detach().numpy())
            real.append(batch['y'].cpu().detach().numpy())

        res = np.vstack(res)
        real = np.vstack(real)


        time = dl.dataset.t
        groups = dl.dataset.groups
        #import pdb
        #pdb.set_trace()
        if self.modifier is not None:
            res,real = self.modifier.inverse_transform(res,real)

        ## BxLxCx3
        if rescaling:
            beauty_string('Scaling back','info',self.verbose)
            if self.normalize_per_group is False:
                for i, c in enumerate(self.target_variables):
                    real[:,:,i] = self.scaler_num[c].inverse_transform(real[:,:,i].reshape(-1,1)).reshape(-1,real.shape[1])
                    for j in range(res.shape[3]):
                        res[:,:,i,j] = self.scaler_num[c].inverse_transform(res[:,:,i,j].reshape(-1,1)).reshape(-1,res.shape[1])
            else:
                for group in np.unique(groups):
                    idx = np.where(groups==group)[0]
                    for i, c in enumerate(self.target_variables):
                        real[idx,:,i] = self.scaler_num[f'{c}_{group}'].inverse_transform(real[idx,:,i].reshape(-1,1)).reshape(-1,real.shape[1])
                        for j in range(res.shape[3]):
                            res[idx,:,i,j] = self.scaler_num[f'{c}_{group}'].inverse_transform(res[idx,:,i,j].reshape(-1,1)).reshape(-1,res.shape[1])

        if self.model.use_quantiles:
            time = pd.DataFrame(time,columns=[i+1 for i in range(res.shape[1])])
            
            if self.group is not None:
                time[self.group] = groups
                time = time.melt(id_vars=[self.group])
            else:
                time = time.melt()
            time.rename(columns={'value':'time','variable':'lag'},inplace=True)

                
            tot = [time]
            for i, c in enumerate(self.target_variables):
                tot.append(pd.DataFrame(real[:,:,i],columns=[i+1 for i in range(res.shape[1])]).melt().rename(columns={'value':c}).drop(columns=['variable']))
                tot.append(pd.DataFrame(res[:,:,i,0],columns=[i+1 for i in range(res.shape[1])]).melt().rename(columns={'value':c+'_low'}).drop(columns=['variable']))
                tot.append(pd.DataFrame(res[:,:,i,1],columns=[i+1 for i in range(res.shape[1])]).melt().rename(columns={'value':c+'_median'}).drop(columns=['variable']))
                tot.append(pd.DataFrame(res[:,:,i,2],columns=[i+1 for i in range(res.shape[1])]).melt().rename(columns={'value':c+'_high'}).drop(columns=['variable']))

            res = pd.concat(tot,axis=1)
        
            
        ## BxLxCx1
        else:
            time = pd.DataFrame(time,columns=[i+1 for i in range(res.shape[1])])#.melt()

            if self.group is not None:
                time[self.group] = groups

                time = time.melt(id_vars=[self.group])
            else:
                time = time.melt()
            time.rename(columns={'value':'time','variable':'lag'},inplace=True)
                 

            tot = [time]
            for i, c in enumerate(self.target_variables):
                tot.append(pd.DataFrame(real[:,:,i],columns=[i+1 for i in range(res.shape[1])]).melt().rename(columns={'value':c}).drop(columns=['variable']))
                tot.append(pd.DataFrame(res[:,:,i,0],columns=[i+1 for i in range(res.shape[1])]).melt().rename(columns={'value':c+'_pred'}).drop(columns=['variable']))
            res = pd.concat(tot,axis=1)

        res['prediction_time'] = res.apply(lambda x: x.time-self.freq*x.lag, axis=1)
        return res
    def inference(self,batch_size:int=100,
                  num_workers:int=4,
                  split_params:Union[None,dict]=None,
                  rescaling:bool=True,
                  data:pd.DataFrame=None,
                  steps_in_future:int=0,
                  check_holes_and_duplicates:bool=True,
                  is_inference:bool=False)->pd.DataFrame:  ##TODO PUSH THIS ON PTF!
        
        """similar to `inference_on_set`
        only change is split_params that must contain this keys but using the default can be sufficient:
        'past_steps','future_steps','shift','keep_entire_seq_while_shifting','starting_point'
        
        skip_step is set to 1 for convenience (generally you want all the predictions)
        You can set split_params to None and use the standard parameters (at your own risck)
   

        Args:
            batch_size (int, optional): see inference_on_set. Defaults to 100.
            num_workers (int, optional): inference_on_set. Defaults to 4.
            split_params (Union[None,dict], optional): inference_on_set. Defaults to None.
            rescaling (bool, optional): inference_on_set. Defaults to True.
            data (pd.DataFrame, optional): startin dataset. Defaults to None.
            steps_in_future (int, optional): if>0 the dataset is extendend in order to make predictions in the future. Defaults to 0.
            check_holes_and_duplicates (bool, optional): if False the routine does not check for holes or for duplicates, set to False for stacked model. Defaults to True.

        Returns:
            pd.DataFrame: predicted values
        """
        beauty_string('Inference on a custom dataset','block',self.verbose)
        self.check_custom = True ##this is a check for the dataset loading
        ## enlarge the dataset in order to have all the rows needed
        if check_holes_and_duplicates:
            if self.group is None:
                ##freq = pd.to_timedelta(np.diff(data.time).min())
                freq = self.freq #TODO port it into PTF
                beauty_string(f'Detected minumum frequency: {freq}','section',self.verbose)
                ## TODO work on this for consistency
                empty = pd.DataFrame({'time':pd.date_range(data.time.min(),data.time.max()+freq*(steps_in_future+self.split_params['past_steps']+self.split_params['future_steps']),freq=freq)})
  
            else:
                freq = pd.to_timedelta(np.diff(data[data[self.group==data[self.group].unique()[0]]].time).min())
                beauty_string(f'Detected minumum frequency: {freq} supposing constant frequence inside the groups','section',self.verbose)
                _min = data.groupby(self.group).time.min()
                _max = data.groupby(self.group).time.max()
                empty = []
                for c in data[self.group].unique():
                    empty.append(pd.DataFrame({self.group:c,'time':pd.date_range(_min.time[_min[self.group]==c].values[0],_max.time[_max[self.group]==c].values[0]+freq*(steps_in_future+self.split_params['past_steps']+self.split_params['future_steps']),freq=freq)}))
                empty = pd.concat(empty,ignore_index=True)
            dataset = empty.merge(data,how='left')
            #TODO port it into PTF
            for c in self.cat_var:
                self.enrich(dataset, c)
        else:
            dataset = data.copy()
        
        
        if split_params is None:
            split_params = {}
            for c in self.split_params.keys():
                if c in ['past_steps','future_steps','shift','keep_entire_seq_while_shifting','starting_point']:
                    split_params[c] = self.split_params[c]
            split_params['skip_step']=1
            data = self.create_data_loader(dataset,**split_params,is_inference=is_inference)
        else:
            data = self.create_data_loader(data,**split_params,is_inference=is_inference)

        res = self.inference_on_set(batch_size=batch_size,num_workers=num_workers,split_params=None,set='custom',rescaling=rescaling,data=data)
        self.check_custom = False
        return res
        
    def save(self, filename:str)->None:
        """save the timeseries object

        Args:
            filename (str): name of the file
        """
        beauty_string('Saving','block',self.verbose)
        with open(f'{filename}.pkl','wb') as f:
            params =  self.__dict__.copy()
            for k in ['model']:
                if k in params.keys():
                    _ = params.pop(k)
            pickle.dump(params,f)


    def load(self,model:Base, filename:str,load_last:bool=True,dirpath:Union[str,None]=None,weight_path:Union[str, None]=None)->None:
        """ Load a saved model

        Args:
            model (Base): class of the model to load (it will be initiated by pytorch-lightening)
            filename (str): filename of the saved model
            load_last (bool, optional): if true the last checkpoint will be loaded otherwise the best (in the validation set). Defaults to True.
            dirpath (Union[str,None], optional): if None we asssume that the model is loaded from the same pc where it has been trained, otherwise we can pass the dirpath where all the stuff has been saved . Defaults to None.
            weight_path (Union[str, None], optional): if None the standard path will be used. Defaults to None.
        """

            
        
        beauty_string('Loading','block',self.verbose)
        self.modifier = None
        self.check_custom = False
        self.is_trained = True
        with open(filename+'.pkl','rb') as f:
            params = pickle.load(f)
            for p in params:
                setattr(self,p, params[p])    
        if 'verbose' in self.config['model_configs'].keys():
            self.config['model_configs'].pop('verbose')
        self.model = model(**self.config['model_configs'],optim_config = self.config['optim_config'],scheduler_config =self.config['scheduler_config'],verbose=self.verbose )
        
        
        if weight_path is not None:
            tmp_path = weight_path
        else:
            if self.dirpath is not None:
                directory = self.dirpath
            else:
                directory = dirpath

            if load_last:
                
                try:
                    tmp_path = os.path.join(directory,self.checkpoint_file_last.split('/')[-1])
                    beauty_string(f"Loading {tmp_path}",'section',self.verbose)

                except Exception as _:
                    beauty_string('checkpoint_file_last not defined try to load best','section',self.verbose)
                    tmp_path = os.path.join(directory,self.checkpoint_file_best.split('/')[-1])
                    beauty_string(f"Loading {tmp_path}",'section',self.verbose)
            else:
                try:
                    
                    tmp_path = os.path.join(directory,self.checkpoint_file_best.split('/')[-1])
                    beauty_string(f"Loading {tmp_path}",'section',self.verbose)
                except Exception as _:
                    beauty_string('checkpoint_file_best not defined try to load best','section',self.verbose)
                    tmp_path = os.path.join(directory,self.checkpoint_file_last.split('/')[-1])
                    beauty_string(f"Loading {tmp_path}",'section',self.verbose)
        try:
            #with torch.serialization.add_safe_globals([ListConfig]):
            if OLD_PL:
                self.model = self.model.load_from_checkpoint(tmp_path,verbose=self.verbose,)
            else:
                self.model = self.model.__class__.load_from_checkpoint(tmp_path,verbose=self.verbose,)
            self.model.to(torch.device("cuda:0" if torch.cuda.is_available() else "cpu"))

        except Exception as e:
            beauty_string(f'There is a problem loading the weights on file {tmp_path} {e}','section',self.verbose)



# -*-Encoding: utf-8 -*-
"""
Authors:
    Li,Yan (liyan22021121@gmail.com)
"""
from torch import nn


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        m.weight.data.normal_(0.0, 0.2)
    elif classname.find("BatchNorm") != -1:
        m.weight.data.normal_(1.0, 0.2)
        m.bias.data.fill_(0)


class MyConvo2d(nn.Module):
    def __init__(self, input_dim, output_dim, kernel_size,  stride = 1, bias = True):
        super(MyConvo2d, self).__init__()
        self.padding = int((kernel_size - 1)/2)
        self.conv = nn.Conv2d(input_dim, output_dim, kernel_size, stride=1, padding=self.padding, bias = bias)

    def forward(self, input):
        output = self.conv(input)
        return output


class Square(nn.Module):
    def __init__(self):
        super(Square,self).__init__()
        pass
    
    def forward(self,in_vect):
        return in_vect**2


class Swish(nn.Module):
    def __init__(self):
        super(Swish,self).__init__()
        pass
    
    def forward(self,in_vect):
        return in_vect*nn.functional.sigmoid(in_vect)


class MeanPoolConv(nn.Module):
    def __init__(self, input_dim, output_dim, kernel_size):
        super(MeanPoolConv, self).__init__()
        self.conv = MyConvo2d(input_dim, output_dim, kernel_size)

    def forward(self, input):
        output = input
        output = self.conv(output)
        return output


class ConvMeanPool(nn.Module):
    def __init__(self, input_dim, output_dim, kernel_size):
        super(ConvMeanPool, self).__init__()
        self.conv = MyConvo2d(input_dim, output_dim, kernel_size)

    def forward(self, input):
        output = self.conv(input)
        return output


class ResidualBlock(nn.Module):
    def __init__(self, input_dim, output_dim, kernel_size, hw, resample=None, normalize=False,AF=nn.ELU()):
        super(ResidualBlock, self).__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.kernel_size = kernel_size
        self.resample = resample
        self.normalize = normalize
        self.bn1 = None
        self.bn2 = None
        self.relu1 = AF
        self.relu2 = AF
        if resample == 'down':
            self.bn1 = nn.LayerNorm([input_dim, hw, hw])
            self.bn2 = nn.LayerNorm([input_dim, hw, hw])
        elif resample == 'none':
            self.bn1 = nn.LayerNorm([input_dim, hw, hw])
            self.bn2 = nn.LayerNorm([input_dim, hw, hw])

        if resample == 'down':
            self.conv_shortcut = MeanPoolConv(input_dim, output_dim, kernel_size = 1)
            self.conv_1 = MyConvo2d(input_dim, input_dim, kernel_size = kernel_size, bias = False)
            self.conv_2 = ConvMeanPool(input_dim, output_dim, kernel_size = kernel_size)
        elif resample == 'none':
            self.conv_shortcut = MyConvo2d(input_dim, output_dim, kernel_size = 1)
            self.conv_1 = MyConvo2d(input_dim, input_dim, kernel_size = kernel_size, bias = False)
            self.conv_2 = MyConvo2d(input_dim, output_dim, kernel_size = kernel_size)
                    
    def forward(self, input):
        if self.input_dim == self.output_dim and self.resample is None:
                shortcut = input
        else:
            shortcut = self.conv_shortcut(input)
        
        if self.normalize is False:
            output = input
            output = self.relu1(output)
            output = self.conv_1(output)
            output = self.relu2(output)
            output = self.conv_2(output)
        else:
            output = input
            output = self.bn1(output)
            output = self.relu1(output)
            output = self.conv_1(output)
            output = self.bn2(output)
            output = self.relu2(output)
            output = self.conv_2(output)

        return shortcut + output


class Res12_Quadratic(nn.Module):
    def __init__(self,inchan,dim,hw,normalize=False,AF=None):
        super(Res12_Quadratic, self).__init__()
        
        self.hw = hw
        self.dim = dim
        self.inchan = inchan
        self.conv1 = MyConvo2d(inchan,dim, 3)
        self.rb1 = ResidualBlock(dim, 2*dim, 3, int(hw), resample = 'down',normalize=normalize,AF=AF)
        self.rbc1 = ResidualBlock(2*dim, 2*dim, 3, int(hw/2), resample = 'none',normalize=normalize,AF=AF)
        self.rb2 = ResidualBlock(2*dim, 4*dim, 3, int(hw/2), resample = 'down',normalize=normalize,AF=AF)
        self.rbc2 = ResidualBlock(4*dim, 4*dim, 3, int(hw/4), resample = 'none',normalize=normalize,AF=AF)
        self.rb3 = ResidualBlock(4*dim, 8*dim, 3, int(hw/4), resample = 'down',normalize=normalize,AF=AF)
        self.rbc3 = ResidualBlock(8*dim, 8*dim, 3, int(hw/8), resample = 'none',normalize=normalize,AF=AF)
        self.ln1 = nn.Linear(int(hw/8)*int(hw/8)*8*dim, 1)
        self.ln2 = nn.Linear(int(hw/8)*int(hw/8)*8*dim, 1)
        self.lq = nn.Linear(int(hw/8)*int(hw/8)*8*dim, 1)
        self.Square = Square()
                
    def forward(self, x_in):
        output = x_in
        output = self.conv1(output)
        # print(output.shape)
        output = self.rb1(output)
        output = self.rbc1(output)
        output = self.rb2(output)
        output = self.rbc2(output)
        output = self.rb3(output)
        output = self.rbc3(output)
        output = output.view(-1, int(self.hw/8)*int(self.hw/8)*8*self.dim)
        output = self.ln1(output)*self.ln2(output)+self.lq(self.Square(output))
        output = output.view(-1)
        return output           


import torch
import torch.nn as nn
from torch.distributions.normal import Normal
from einops import rearrange
from ..autoformer.layers import series_decomp

class encoder(nn.Module):
    def __init__(self, seq_len,num_experts,hidden_size ):
        super(encoder, self).__init__()
        input_size = seq_len
        num_experts = num_experts
        encoder_hidden_size = hidden_size

        self.distribution_fit = nn.Sequential(nn.Linear(input_size, encoder_hidden_size, bias=False), nn.ReLU(),
                                              nn.Linear(encoder_hidden_size, num_experts, bias=False))

    def forward(self, x):
        mean = torch.mean(x, dim=-1)
        out = self.distribution_fit(mean)
        return out


class RevIN(nn.Module):
    def __init__(self, num_features: int, eps=1e-5, affine=True, subtract_last=False):
        """
        :param num_features: the number of features or channels
        :param eps: a value added for numerical stability
        :param affine: if True, RevIN has learnable affine parameters
        """
        super(RevIN, self).__init__()
        self.num_features = num_features
        self.eps = eps
        self.affine = affine
        self.subtract_last = subtract_last
        if self.affine:
            self._init_params()

    def forward(self, x, mode: str):
        if mode == 'norm':
            self._get_statistics(x)
            x = self._normalize(x)
        elif mode == 'denorm':
            x = self._denormalize(x)
        else:
            raise NotImplementedError
        return x

    def _init_params(self):
        # initialize RevIN params: (C,)
        self.affine_weight = nn.Parameter(torch.ones(self.num_features))
        self.affine_bias = nn.Parameter(torch.zeros(self.num_features))

    def _get_statistics(self, x):
        dim2reduce = tuple(range(1, x.ndim - 1))
        if self.subtract_last:
            self.last = x[:, -1, :].unsqueeze(1)
        else:
            self.mean = torch.mean(x, dim=dim2reduce, keepdim=True).detach()  
        self.stdev = torch.sqrt(torch.var(x, dim=dim2reduce, keepdim=True, unbiased=False) + self.eps).detach()

    def _normalize(self, x):
        if self.subtract_last:
            x = x - self.last
        else:
            x = x - self.mean
        x = x / self.stdev
        if self.affine:
            x = x * self.affine_weight
            x = x + self.affine_bias
        return x

    def _denormalize(self, x):
        if self.affine:
            x = x - self.affine_bias
            x = x / (self.affine_weight + self.eps * self.eps)
        x = x * self.stdev
        if self.subtract_last:
            x = x + self.last
        else:
            x = x + self.mean
        return x


class Linear_extractor(nn.Module):
    """
    Paper link: https://arxiv.org/pdf/2205.13504.pdf
    """

    def __init__(self, seq_len,d_model,enc_in,CI,moving_avg, individual=False):
        """
        individual: Bool, whether shared model among different variates.
        """
        super(Linear_extractor, self).__init__()
        self.seq_len = seq_len

        self.pred_len = d_model
        self.decompsition = series_decomp(moving_avg)
        self.individual = individual
        self.channels = enc_in
        self.enc_in = 1 if CI else enc_in
        if self.individual:
            self.Linear_Seasonal = nn.ModuleList()
            self.Linear_Trend = nn.ModuleList()

            for i in range(self.channels):
                self.Linear_Seasonal.append(
                    nn.Linear(self.seq_len, self.pred_len))
                self.Linear_Trend.append(
                    nn.Linear(self.seq_len, self.pred_len))

                self.Linear_Seasonal[i].weight = nn.Parameter(
                    (1 / self.seq_len) * torch.ones([self.pred_len, self.seq_len]))
                self.Linear_Trend[i].weight = nn.Parameter(
                    (1 / self.seq_len) * torch.ones([self.pred_len, self.seq_len]))
        else:
            self.Linear_Seasonal = nn.Linear(self.seq_len, self.pred_len)
            self.Linear_Trend = nn.Linear(self.seq_len, self.pred_len)

            self.Linear_Seasonal.weight = nn.Parameter(
                (1 / self.seq_len) * torch.ones([self.pred_len, self.seq_len]))
            self.Linear_Trend.weight = nn.Parameter(
                (1 / self.seq_len) * torch.ones([self.pred_len, self.seq_len]))



    def encoder(self, x):
        seasonal_init, trend_init = self.decompsition(x)
        seasonal_init, trend_init = seasonal_init.permute(
            0, 2, 1), trend_init.permute(0, 2, 1)
        if self.individual:
            seasonal_output = torch.zeros([seasonal_init.size(0), seasonal_init.size(1), self.pred_len],
                                          dtype=seasonal_init.dtype).to(seasonal_init.device)
            trend_output = torch.zeros([trend_init.size(0), trend_init.size(1), self.pred_len],
                                       dtype=trend_init.dtype).to(trend_init.device)
            for i in range(self.channels):
                seasonal_output[:, i, :] = self.Linear_Seasonal[i](
                    seasonal_init[:, i, :])
                trend_output[:, i, :] = self.Linear_Trend[i](
                    trend_init[:, i, :])
        else:
            seasonal_output = self.Linear_Seasonal(seasonal_init)
            trend_output = self.Linear_Trend(trend_init)
        x = seasonal_output + trend_output
        return x.permute(0, 2, 1)

    def forecast(self, x_enc):
        # Encoder
        return self.encoder(x_enc)


    def forward(self, x_enc):
        if x_enc.shape[0] == 0:
            return torch.empty((0, self.pred_len, self.enc_in)).to(x_enc.device)
        dec_out = self.forecast(x_enc)
        return dec_out[:, -self.pred_len:, :]  # [B, L, D]


class SparseDispatcher(object):
    """Helper for implementing a mixture of experts.
    The purpose of this class is to create input mini-batches for the
    experts and to combine the results of the experts to form a unified
    output tensor.
    There are two functions:
    dispatch - take an input Tensor and create input Tensors for each expert.
    combine - take output Tensors from each expert and form a combined output
      Tensor.  Outputs from different experts for the same batch element are
      summed together, weighted by the provided "gates".
    The class is initialized with a "gates" Tensor, which specifies which
    batch elements go to which experts, and the weights to use when combining
    the outputs.  Batch element b is sent to expert e iff gates[b, e] != 0.
    The inputs and outputs are all two-dimensional [batch, depth].
    Caller is responsible for collapsing additional dimensions prior to
    calling this class and reshaping the output to the original shape.
    See common_layers.reshape_like().
    Example use:
    gates: a float32 `Tensor` with shape `[batch_size, num_experts]`
    inputs: a float32 `Tensor` with shape `[batch_size, input_size]`
    experts: a list of length `num_experts` containing sub-networks.
    dispatcher = SparseDispatcher(num_experts, gates)
    expert_inputs = dispatcher.dispatch(inputs)
    expert_outputs = [experts[i](expert_inputs[i]) for i in range(num_experts)]
    outputs = dispatcher.combine(expert_outputs)
    The preceding code sets the output for a particular example b to:
    output[b] = Sum_i(gates[b, i] * experts[i](inputs[b]))
    This class takes advantage of sparsity in the gate matrix by including in the
    `Tensor`s for expert i only the batch elements for which `gates[b, i] > 0`.
    """

    def __init__(self, num_experts, gates):
        """Create a SparseDispatcher."""

        self._gates = gates
        self._num_experts = num_experts
        # sort experts
        sorted_experts, index_sorted_experts = torch.nonzero(gates).sort(0)
        # drop indices
        _, self._expert_index = sorted_experts.split(1, dim=1)
        # get according batch index for each expert
        self._batch_index = torch.nonzero(gates)[index_sorted_experts[:, 1], 0]
        # calculate num samples that each expert gets
        self._part_sizes = (gates > 0).sum(0).tolist()
        # expand gates to match with self._batch_index
        gates_exp = gates[self._batch_index.flatten()]
        self._nonzero_gates = torch.gather(gates_exp, 1, self._expert_index)

    def dispatch(self, inp):
        """Create one input Tensor for each expert.
        The `Tensor` for an expert `i` contains the slices of `inp` corresponding
        to the batch elements `b` where `gates[b, i] > 0`.
        Args:
          inp: a `Tensor` of shape "[batch_size, <extra_input_dims>]`
        Returns:
          a list of `num_experts` `Tensor`s with shapes
            `[expert_batch_size_i, <extra_input_dims>]`.
        """

        # assigns samples to experts whose gate is nonzero

        # expand according to batch index so we can just split by _part_sizes
        inp_exp = inp[self._batch_index].squeeze(1)
        return torch.split(inp_exp, self._part_sizes, dim=0)
 
    def combine(self, expert_out, multiply_by_gates=True):
        """Sum together the expert output, weighted by the gates.
        The slice corresponding to a particular batch element `b` is computed
        as the sum over all experts `i` of the expert output, weighted by the
        corresponding gate values.  If `multiply_by_gates` is set to False, the
        gate values are ignored.
        Args:
          expert_out: a list of `num_experts` `Tensor`s, each with shape
            `[expert_batch_size_i, <extra_output_dims>]`.
          multiply_by_gates: a boolean
        Returns:
          a `Tensor` with shape `[batch_size, <extra_output_dims>]`.
        """
        # apply exp to expert outputs, so we are not longer in log space
  
        stitched = torch.cat(expert_out, 0)

        if multiply_by_gates:
            # stitched = stitched.mul(self._nonzero_gates)
            stitched = torch.einsum("i...,ij->i...", stitched, self._nonzero_gates)

        shape = list(expert_out[-1].shape)
        shape[0] = self._gates.size(0)
        zeros = torch.zeros(*shape, requires_grad=True, device=stitched.device)
        # combine samples that have been processed by the same k experts
        combined = zeros.index_add(0, self._batch_index, stitched.float())
        return combined

    def expert_to_gates(self):
        """Gate values corresponding to the examples in the per-expert `Tensor`s.
        Returns:
          a list of `num_experts` one-dimensional `Tensor`s with type `tf.float32`
              and shapes `[expert_batch_size_i]`
        """
        # split nonzero gates for each expert
        return torch.split(self._nonzero_gates, self._part_sizes, dim=0)


class Linear_extractor_cluster(nn.Module):
    """Call a Sparsely gated mixture of experts layer with 1-layer Feed-Forward networks as experts.
    Args:
    input_size: integer - size of the input
    output_size: integer - size of the input
    num_experts: an integer - number of experts
    hidden_size: an integer - hidden size of the experts
    noisy_gating: a boolean
    k: an integer - how many experts to use for each batch element
    """

    def __init__(self, noisy_gating,num_experts,seq_len,k,d_model,enc_in,CI,moving_avg,hidden_size ):
        super(Linear_extractor_cluster, self).__init__()
        self.noisy_gating = noisy_gating
        self.num_experts = num_experts
        self.input_size = seq_len
        self.k = k
        # instantiate experts
        self.experts = nn.ModuleList([Linear_extractor(seq_len,d_model,enc_in,CI,moving_avg) for _ in range(self.num_experts)])
        self.W_h = nn.Parameter(torch.eye(self.num_experts))
        self.gate = encoder(seq_len,num_experts,hidden_size)
        self.noise = encoder(seq_len,num_experts,hidden_size)

        self.n_vars = enc_in
        self.revin = RevIN(self.n_vars)

        self.CI = CI
        self.softplus = nn.Softplus()
        self.softmax = nn.Softmax(1)
        self.register_buffer("mean", torch.tensor([0.0]))
        self.register_buffer("std", torch.tensor([1.0]))
        assert self.k <= self.num_experts

    def cv_squared(self, x):
        """The squared coefficient of variation of a sample.
        Useful as a loss to encourage a positive distribution to be more uniform.
        Epsilons added for numerical stability.
        Returns 0 for an empty Tensor.
        Args:
        x: a `Tensor`.
        Returns:
        a `Scalar`.
        """
        eps = 1e-10
        # if only num_experts = 1

        if x.shape[0] == 1:
            return torch.tensor([0], device=x.device, dtype=x.dtype)
        return x.float().var() / (x.float().mean() ** 2 + eps)

    def _gates_to_load(self, gates):
        """Compute the true load per expert, given the gates.
        The load is the number of examples for which the corresponding gate is >0.
        Args:
        gates: a `Tensor` of shape [batch_size, n]
        Returns:
        a float32 `Tensor` of shape [n]
        """
        return (gates > 0).sum(0)

    def _prob_in_top_k(
        self, clean_values, noisy_values, noise_stddev, noisy_top_values
    ):
        """Helper function to NoisyTopKGating.
        Computes the probability that value is in top k, given different random noise.
        This gives us a way of backpropagating from a loss that balances the number
        of times each expert is in the top k experts per example.
        In the case of no noise, pass in None for noise_stddev, and the result will
        not be differentiable.
        Args:
        clean_values: a `Tensor` of shape [batch, n].
        noisy_values: a `Tensor` of shape [batch, n].  Equal to clean values plus
          normally distributed noise with standard deviation noise_stddev.
        noise_stddev: a `Tensor` of shape [batch, n], or None
        noisy_top_values: a `Tensor` of shape [batch, m].
           "values" Output of tf.top_k(noisy_top_values, m).  m >= k+1
        Returns:
        a `Tensor` of shape [batch, n].
        """
        batch = clean_values.size(0)
        m = noisy_top_values.size(1)
        top_values_flat = noisy_top_values.flatten()

        threshold_positions_if_in = (
            torch.arange(batch, device=clean_values.device) * m + self.k
        )
        threshold_if_in = torch.unsqueeze(
            torch.gather(top_values_flat, 0, threshold_positions_if_in), 1
        )
        is_in = torch.gt(noisy_values, threshold_if_in)
        threshold_positions_if_out = threshold_positions_if_in - 1
        threshold_if_out = torch.unsqueeze(
            torch.gather(top_values_flat, 0, threshold_positions_if_out), 1
        )
        # is each value currently in the top k.
        normal = Normal(self.mean, self.std)
        prob_if_in = normal.cdf((clean_values - threshold_if_in) / noise_stddev)
        prob_if_out = normal.cdf((clean_values - threshold_if_out) / noise_stddev)
        prob = torch.where(is_in, prob_if_in, prob_if_out)
        return prob

    def noisy_top_k_gating(self, x, train, noise_epsilon=1e-2):
        """Noisy top-k gating.
        See paper: https://arxiv.org/abs/1701.06538.
        Args:
          x: input Tensor with shape [batch_size, input_size]
          train: a boolean - we only add noise at training time.
          noise_epsilon: a float
        Returns:
          gates: a Tensor with shape [batch_size, num_experts]
          load: a Tensor with shape [num_experts]
        """
        clean_logits = self.gate(x)

        if self.noisy_gating and train:
            raw_noise_stddev = self.noise(x)
            noise_stddev = self.softplus(raw_noise_stddev) + noise_epsilon
            noise = torch.randn_like(clean_logits)
            noisy_logits = clean_logits + (noise * noise_stddev)
            logits = noisy_logits @ self.W_h
        else:
            logits = clean_logits

        # calculate topk + 1 that will be needed for the noisy gates
        logits = self.softmax(logits)
        top_logits, top_indices = logits.topk(min(self.k + 1, self.num_experts), dim=1)
        top_k_logits = top_logits[:, : self.k]
        top_k_indices = top_indices[:, : self.k]
        top_k_gates = top_k_logits / (
            top_k_logits.sum(1, keepdim=True) + 1e-6
        )  # normalization

        zeros = torch.zeros_like(logits, requires_grad=True)
        gates = zeros.scatter(1, top_k_indices, top_k_gates)

        if self.noisy_gating and self.k < self.num_experts and train:
            load = (
                self._prob_in_top_k(
                    clean_logits, noisy_logits, noise_stddev, top_logits
                )
            ).sum(0)
        else:
            load = self._gates_to_load(gates)
        return gates, load

    def forward(self, x, loss_coef=1):
        """Args:
        x: tensor shape [batch_size, input_size]
        train: a boolean scalar.
        loss_coef: a scalar - multiplier on load-balancing losses

        Returns:
        y: a tensor with shape [batch_size, output_size].
        extra_training_loss: a scalar.  This should be added into the overall
        training loss of the model.  The backpropagation of this loss
        encourages all experts to be approximately equally used across a batch.
        """
        gates, load = self.noisy_top_k_gating(x, self.training)
        # calculate importance loss
        importance = gates.sum(0)
        loss = self.cv_squared(importance) + self.cv_squared(load)
        loss *= loss_coef

        dispatcher = SparseDispatcher(self.num_experts, gates)
        if self.CI:
            x_norm = rearrange(x, "(x y) l c -> x l (y c)", y=self.n_vars)
            x_norm = self.revin(x_norm, "norm")
            x_norm = rearrange(x_norm, "x l (y c) -> (x y) l c", y=self.n_vars)
        else:
            x_norm = self.revin(x, "norm")

        expert_inputs = dispatcher.dispatch(x_norm)

        gates = dispatcher.expert_to_gates()

        expert_outputs = [
            self.experts[i](expert_inputs[i]) for i in range(self.num_experts)
        ]
        #y = dispatcher.combine([e for e in expert_outputs if len(e)>0])
        #with torch._dynamo.disable():
        y = dispatcher.combine(expert_outputs)
        return y, loss


## Copyright 2022 DLinear Authors (https://github.com/cure-lab/LTSF-Linear/tree/main?tab=Apache-2.0-1-ov-file#readme)
## Code modified for align the notation and the batch generation
## extended to all present in informer, autoformer folder


from torch import  nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import List,Union
from ..data_structure.utils import beauty_string
from .utils import  get_activation,get_scope,QuantileLossMO
from .autoformer.layers import AutoCorrelation, AutoCorrelationLayer, Encoder, Decoder,\
    EncoderLayer, DecoderLayer, my_Layernorm, series_decomp,PositionalEmbedding
from .utils import Embedding_cat_variables


  
class Autoformer(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss= True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 label_len: int, 
                 d_model:int,
                 dropout_rate:float,
                 kernel_size:int,
                 activation:str='torch.nn.ReLU',
                 factor: float=0.5,
                 n_head:int=1,
                 n_layer_encoder:int=2,
                 n_layer_decoder:int=2,
                 hidden_size:int=1048,
                 **kwargs
                )->None:
        """Autoformer from https://github.com/cure-lab/LTSF-Linear

        Args:
            label_len (int): see the original implementation, seems like a warmup dimension (the decoder part will produce also some past predictions that are filter out at the end)
            d_model (int): embedding dimension of the attention layer
            dropout_rate (float): dropout raye
            kernel_size (int): kernel size
            activation (str, optional): _description_. Defaults to 'torch.nn.ReLU'.
            factor (int, optional): parameter of `.autoformer.layers.AutoCorrelation` for find the top k. Defaults to 0.5.
            n_head (int, optional): number of heads. Defaults to 1.
            n_layer_encoder (int, optional): number of  encoder layers. Defaults to 2.
            n_layer_decoder (int, optional): number of decoder layers. Defaults to 2.
            hidden_size (int, optional): output dimension of the transformer layer. Defaults to 1048.
        """
        super().__init__(**kwargs)
        beauty_string(self.description,'info',True)

        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
        if isinstance(activation,str):
            activation = get_activation(activation)
        else:
            beauty_string('There is a bug in pytorch lightening, the constructior is called twice ','info',self.verbose)
        
   
        
        self.save_hyperparameters(logger=False)
      


        
        self.seq_len = self.past_steps
        self.label_len = label_len
        self.pred_len = self.future_steps

        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps+label_len,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels


        # Decomp
        self.decomp = series_decomp(kernel_size)



        self.linear_encoder = nn.Sequential(nn.Linear(self.past_channels+emb_past_out_channel,self.past_channels*2),
                                            activation(),
                                            nn.Dropout(dropout_rate),
                                            nn.Linear(self.past_channels*2,d_model*2),
                                            activation(),
                                            nn.Dropout(dropout_rate),
                                            nn.Linear(d_model*2,d_model))
        
        self.linear_decoder = nn.Sequential(nn.Linear(self.future_channels+emb_fut_out_channel,self.future_channels*2),
                                            activation(),
                                            nn.Dropout(dropout_rate),
                                            nn.Linear(self.future_channels*2,d_model*2),
                                            activation() ,nn.Dropout(dropout_rate),
                                            nn.Linear(d_model*2,d_model))
       
        #self.final_layer =  nn.Linear(self.past_channels,self.out_channels)
       
        # Encoder
        self.encoder = Encoder(
            [
                EncoderLayer(
                    AutoCorrelationLayer(
                        AutoCorrelation(False, factor, attention_dropout=dropout_rate,
                                        output_attention=False),
                        d_model, n_head),
                    d_model,
                    hidden_size,
                    moving_avg=kernel_size,
                    dropout=dropout_rate,
                    activation=activation
                ) for _ in range(n_layer_encoder)
            ],
            norm_layer=my_Layernorm(d_model)
        )
        # Decoder
        self.decoder = Decoder(
            [
                DecoderLayer(
                    AutoCorrelationLayer(
                        AutoCorrelation(True, factor, attention_dropout=dropout_rate,
                                        output_attention=False),
                        d_model, n_head),
                    AutoCorrelationLayer(
                        AutoCorrelation(False, factor, attention_dropout=dropout_rate,
                                        output_attention=False),
                        d_model, n_head),
                    d_model,
                    self.out_channels,
                    hidden_size,
                    moving_avg=kernel_size,
                    dropout=dropout_rate,
                    activation=activation,
                )
                for _ in range(n_layer_decoder)
            ],
            norm_layer=my_Layernorm(d_model),
            projection=nn.Linear(d_model, self.out_channels*self.mul, bias=True)
        )
        self.projection = nn.Linear(self.past_channels,self.out_channels*self.mul )
    def can_be_compiled(self):
        return True
    def forward(self, batch):
        


        idx_target_future = batch['idx_target_future'][0]
        x = batch['x_num_past'].to(self.device)
        BS = x.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
        
  
        if 'x_num_future' in batch.keys():
            x_future = batch['x_num_future'].to(self.device)
            x_future[:,-self.pred_len:,idx_target_future] = 0
        
        

        
        mean = torch.mean(x, dim=1).unsqueeze(1).repeat(1, self.pred_len, 1)

        zeros = torch.zeros([x_future.shape[0], self.pred_len, x.shape[2]], device=x.device)
        seasonal_init, trend_init = self.decomp(x)
        # decoder input
        trend_init = torch.cat([trend_init[:, -self.label_len:, :], mean], dim=1)
        seasonal_init = torch.cat([seasonal_init[:, -self.label_len:, :], zeros], dim=1)
        # enc
        enc_out = self.linear_encoder(torch.cat([x,emb_past],2))
        enc_out, attns = self.encoder(enc_out, attn_mask=None)
        # dec
        dec_out = self.linear_decoder(torch.cat([x_future,emb_fut],2))
        seasonal_part, trend_part = self.decoder(dec_out, enc_out, x_mask=None, cross_mask=None, trend=trend_init)
        # final

        trend_part = self.projection(trend_part)
        dec_out = trend_part + seasonal_part

    
        BS = dec_out.shape[0]
        
        return dec_out[:, -self.pred_len:, :].reshape(BS,self.pred_len,-1,self.mul)  # [B, L, D,MUL]
         
        



import torch
import torch.nn as nn


import numpy as np

from math import sqrt


class TriangularCausalMask():
    def __init__(self, B, L,device):
        mask_shape = [B, 1, L, L]
        with torch.no_grad():
            self._mask = torch.triu(torch.ones(mask_shape, dtype=torch.bool), diagonal=1).to(device)

    @property
    def mask(self):
        return self._mask

class ProbMask():
    def __init__(self, B, H, L, index, scores,device):
        _mask = torch.ones(L, scores.shape[-1], dtype=torch.bool).triu(1).to(device)
        _mask_ex = _mask[None, None, :].expand(B, H, L, scores.shape[-1])
        indicator = _mask_ex[torch.arange(B)[:, None, None],torch.arange(H)[None, :, None],index, :]
        self._mask = indicator.view(scores.shape)
    
    @property
    def mask(self):
        return self._mask   


class FullAttention(nn.Module):
    def __init__(self, mask_flag=True, factor=5, scale=None, attention_dropout=0.1, output_attention=False):
        super(FullAttention, self).__init__()
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)
        
    def forward(self, queries, keys, values, attn_mask):
        B, L, H, E = queries.shape
        _, S, _, D = values.shape
        scale = self.scale or 1./sqrt(E)

        scores = torch.einsum("blhe,bshe->bhls", queries, keys)
        if self.mask_flag:
            if attn_mask is None:
                attn_mask = TriangularCausalMask(B, L, device=queries.device)

            scores.masked_fill_(attn_mask.mask, -np.inf)

        A = self.dropout(torch.softmax(scale * scores, dim=-1))
        V = torch.einsum("bhls,bshd->blhd", A, values)
        
        if self.output_attention:
            return (V.contiguous(), A)
        else:
            return (V.contiguous(), None)

class ProbAttention(nn.Module):
    def __init__(self, mask_flag=True, factor=5, scale=None, attention_dropout=0.1, output_attention=False):
        super(ProbAttention, self).__init__()
        self.factor = factor
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)

    def _prob_QK(self, Q, K, sample_k, n_top): # n_top: c*ln(L_q)
        # Q [B, H, L, D]
        B, H, L_K, E = K.shape
        _, _, L_Q, _ = Q.shape

        # calculate the sampled Q_K
        K_expand = K.unsqueeze(-3).expand(B, H, L_Q, L_K, E)
        index_sample = torch.randint(L_K, (L_Q, sample_k)) # real U = U_part(factor*ln(L_k))*L_q
        K_sample = K_expand[:, :, torch.arange(L_Q).unsqueeze(1), index_sample, :]
        Q_K_sample = torch.matmul(Q.unsqueeze(-2), K_sample.transpose(-2, -1)).squeeze(-2)

        # find the Top_k query with sparisty measurement
        M = Q_K_sample.max(-1)[0] - torch.div(Q_K_sample.sum(-1), L_K)
        M_top = M.topk(n_top, sorted=False)[1]

        # use the reduced Q to calculate Q_K
        Q_reduce = Q[torch.arange(B)[:, None, None],
                     torch.arange(H)[None, :, None],
                     M_top, :] # factor*ln(L_q)
        Q_K = torch.matmul(Q_reduce, K.transpose(-2, -1)) # factor*ln(L_q)*L_k

        return Q_K, M_top

    def _get_initial_context(self, V, L_Q):
        B, H, L_V, D = V.shape
        if not self.mask_flag:
            # V_sum = V.sum(dim=-2)
            V_sum = V.mean(dim=-2)
            contex = V_sum.unsqueeze(-2).expand(B, H, L_Q, V_sum.shape[-1]).clone()
        else: # use mask
            assert(L_Q == L_V) # requires that L_Q == L_V, i.e. for self-attention only
            contex = V.cumsum(dim=-2)
        return contex

    def _update_context(self, context_in, V, scores, index, L_Q, attn_mask):
        B, H, L_V, D = V.shape

        if self.mask_flag:
            attn_mask = ProbMask(B, H, L_Q, index, scores, device=V.device)
            scores.masked_fill_(attn_mask.mask, -np.inf)

        attn = torch.softmax(scores, dim=-1) # nn.Softmax(dim=-1)(scores)

        context_in[torch.arange(B)[:, None, None],
                   torch.arange(H)[None, :, None],
                   index, :] = torch.matmul(attn, V).type_as(context_in)
        if self.output_attention:
            attns = (torch.ones([B, H, L_V, L_V])/L_V).type_as(attn).to(attn.device)
            attns[torch.arange(B)[:, None, None], torch.arange(H)[None, :, None], index, :] = attn
            return (context_in, attns)
        else:
            return (context_in, None)

    def forward(self, queries, keys, values, attn_mask):
        B, L_Q, H, D = queries.shape
        _, L_K, _, _ = keys.shape

        queries = queries.transpose(2,1)
        keys = keys.transpose(2,1)
        values = values.transpose(2,1)

        U_part = self.factor * np.ceil(np.log(L_K)).astype('int').item() # c*ln(L_k)
        u = self.factor * np.ceil(np.log(L_Q)).astype('int').item() # c*ln(L_q) 

        U_part = U_part if U_part<L_K else L_K
        u = u if u<L_Q else L_Q
        
        scores_top, index = self._prob_QK(queries, keys, sample_k=U_part, n_top=u) 

        # add scale factor
        scale = self.scale or 1./sqrt(D)
        if scale is not None:
            scores_top = scores_top * scale
        # get the context
        context = self._get_initial_context(values, L_Q)
        # update the context with selected top_k queries
        context, attn = self._update_context(context, values, scores_top, index, L_Q, attn_mask)
        
        return context.transpose(2,1).contiguous(), attn


class AttentionLayer(nn.Module):
    def __init__(self, attention, d_model, n_heads, 
                 d_keys=None, d_values=None, mix=False):
        super(AttentionLayer, self).__init__()

        d_keys = d_keys or (d_model//n_heads)
        d_values = d_values or (d_model//n_heads)

        self.inner_attention = attention
        self.query_projection = nn.Linear(d_model, d_keys * n_heads)
        self.key_projection = nn.Linear(d_model, d_keys * n_heads)
        self.value_projection = nn.Linear(d_model, d_values * n_heads)
        self.out_projection = nn.Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads
        self.mix = mix

    def forward(self, queries, keys, values, attn_mask):
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        H = self.n_heads

        queries = self.query_projection(queries).view(B, L, H, -1)
        keys = self.key_projection(keys).view(B, S, H, -1)
        values = self.value_projection(values).view(B, S, H, -1)

        out, attn = self.inner_attention(
            queries,
            keys,
            values,
            attn_mask
        )
        if self.mix:
            out = out.transpose(2,1).contiguous()
        out = out.view(B, L, -1)

        return self.out_projection(out), attn



## Copyright 2023 Yunhao Zhang and Junchi Yan (https://github.com/Thinklab-SJTU/Crossformer?tab=Apache-2.0-1-ov-file#readme)
## Code modified for align the notation and the batch generation
## extended to all present in crossformer folder



from torch import  nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import List,Union
from einops import  repeat
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .crossformer.cross_encoder import Encoder
from .crossformer.cross_decoder import Decoder
from .crossformer.cross_embed import DSW_embedding
from .utils import Embedding_cat_variables
from math import ceil
from .utils import Embedding_cat_variables
from .utils import get_activation

  

  
class CrossFormer(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True

    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    def __init__(self, 
                 
                 d_model:int,
                 hidden_size:int,
                 n_head:int,
                 seg_len:int,
                 n_layer_encoder:int,
                 win_size:int,
                 factor:int=10,
                 dropout_rate:float=0.1,
                 activation:str='torch.nn.ReLU',

                 **kwargs)->None:
        """  CrossFormer (https://openreview.net/forum?id=vSVLM2j9eie)

        
        Args:
            d_model (int): The dimensionality of the model.
            hidden_size (int): The size of the hidden layers.
            n_head (int): The number of attention heads.
            seg_len (int): The length of the segments.
            n_layer_encoder (int): The number of layers in the encoder.
            win_size (int): The size of the window for attention.
            factor (int, optional): see .crossformer.attn.TwoStageAttentionLayer. Defaults to 10.
            dropout_rate (float, optional): The dropout rate. Defaults to 0.1.
            activation (str, optional): The activation function to use. Defaults to 'torch.nn.ReLU'.
            **kwargs: Additional keyword arguments for the parent class.
        
        Returns:
            None: This method does not return a value.
        
        Raises:
            ValueError: If the activation function is not recognized.
            

        """
      
        if isinstance(activation, str):
            activation = get_activation(activation)
        super().__init__(**kwargs)
        self.save_hyperparameters(logger=False)
        

        # The padding operation to handle invisible sgemnet length
        self.pad_past_steps = ceil(1.0 *self.past_steps / seg_len) * seg_len
        self.pad_future_steps = ceil(1.0 * self.future_steps / seg_len) * seg_len
        self.past_steps_add = self.pad_past_steps - self.past_steps

        # Embedding
        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels
        self.enc_value_embedding = DSW_embedding(seg_len, d_model)
        self.enc_pos_embedding = nn.Parameter(torch.randn(1, self.past_channels+emb_past_out_channel, (self.pad_past_steps // seg_len), d_model))
        self.pre_norm = nn.LayerNorm(d_model)

        ## Custom embeddings ##these are not used in crossformer


        self.encoder = Encoder(n_layer_encoder, win_size, d_model, n_head, hidden_size, block_depth = 1, \
                                    dropout = dropout_rate,in_seg_num = (self.pad_past_steps // seg_len), factor = factor)
        
        # Decoder
        self.dec_pos_embedding = nn.Parameter(torch.randn(1, self.past_channels+emb_past_out_channel, (self.pad_future_steps // seg_len), d_model))
        self.decoder = Decoder(seg_len, n_layer_encoder + 1, d_model, n_head, hidden_size, dropout_rate, \
                                    out_seg_num = (self.pad_future_steps // seg_len), factor = factor)
     
        dim = self.past_channels+emb_past_out_channel+emb_fut_out_channel+self.future_channels
        self.final_layer = nn.Sequential(activation(),
                                         nn.Linear(dim, dim//2),
                                         activation(),
                                         nn.Linear(dim//2, self.mul*self.out_channels ))
        
                                         
        
    def can_be_compiled(self):
        return True
        
    def forward(self, batch):

        x_seq = batch['x_num_past'].to(self.device)#[:,:,idx_target]
        BS = x_seq.shape[0]
        
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
            
        tmp_future = [emb_fut]
        
        if 'x_num_future' in batch.keys():
            x_future = batch['x_num_future'].to(self.device)
            tmp_future.append(x_future)

        x_seq = torch.cat([x_seq,emb_past],2) 
        batch_size = x_seq.shape[0]
        if (self.past_steps_add != 0):
            x_seq = torch.cat((x_seq[:, :1, :].expand(-1, self.past_steps_add, -1), x_seq), dim = 1)
        x_seq = self.enc_value_embedding(x_seq)
        x_seq += self.enc_pos_embedding
        x_seq = self.pre_norm(x_seq)        
        enc_out = self.encoder(x_seq)
        dec_in = repeat(self.dec_pos_embedding, 'b ts_d l d -> (repeat b) ts_d l d', repeat = batch_size)
        predict_y = self.decoder(dec_in, enc_out)
        res = predict_y[:, :self.future_steps,:]
        tmp_future.append(res)
        res = self.final_layer(torch.cat(tmp_future,2))
        return res.reshape(BS, -1, self.out_channels,self.mul)



## Copyright 2022 DLinear Authors (https://github.com/cure-lab/LTSF-Linear/tree/main?tab=Apache-2.0-1-ov-file#readme)
## Code modified for align the notation and the batch generation
## extended to all present in informer, autoformer folder

from torch import nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from .utils import QuantileLossMO, get_activation
from typing import List, Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import Embedding_cat_variables
    


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """
    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """
    Series decomposition block
    """
    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class LinearTS(Base):
    handle_multivariate = True
    handle_future_covariates = True
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    description+='\n THE SIMPLE IMPLEMENTATION DOES NOT USE CATEGORICAL NOR FUTURE VARIABLES'
    
    def __init__(self, 
               
                 kernel_size:int,
                 hidden_size:int,
                 dropout_rate:float=0.1,
                 activation:str='torch.nn.ReLU',
                 kind:str='linear',
                 use_bn:bool=False,
                 simple:bool=False,
                 **kwargs)->None:
        """Initialize the model with specified parameters. Linear model from https://github.com/cure-lab/LTSF-Linear/blob/main/run_longExp.py
        
        Args:
            kernel_size (int): Kernel dimension for the initial moving average.
            hidden_size (int): Hidden size of the linear block.
            dropout_rate (float, optional): Dropout rate in Dropout layers. Default is 0.1.
            activation (str, optional): Activation function in PyTorch. Default is 'torch.nn.ReLU'.
            kind (str, optional): Type of model, can be 'linear', 'dlinear' (de-trending), or 'nlinear' (differential). Defaults to 'linear'.
            use_bn (bool, optional): If True, Batch Normalization layers will be added and Dropouts will be removed. Default is False.
            simple (bool, optional): If True, the model used is the same as illustrated in the paper; otherwise, a more complex model with the same idea is used. Default is False.
            **kwargs: Additional keyword arguments for the parent class.
        
        Raises:
            ValueError: If an invalid activation function is provided.
        """
        
        super().__init__(**kwargs)

        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
            use_bn = False
            
        if isinstance(activation, str):
            activation = get_activation(activation)
        else:
            beauty_string('There is a bug in pytorch lightening, the constructior is called twice','info',self.verbose)
        
        self.save_hyperparameters(logger=False)
      
        self.kind = kind
       

        self.simple = simple
       
        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels



        ## ne faccio uno per ogni canale
        self.linear =  nn.ModuleList()

            
        if kind=='dlinear':
            self.decompsition = series_decomp(kernel_size)    
            self.Linear_Trend = nn.ModuleList()
            for _ in range(self.out_channels):
                self.Linear_Trend.append(nn.Linear(self.past_steps,self.future_steps))
            
        
        for _ in range(self.out_channels):
            if simple:
                self.linear.append(nn.Linear(self.past_steps,self.future_steps*self.mul))
                                               
            else:
                self.linear.append(nn.Sequential(nn.Linear(emb_past_out_channel*self.past_steps+emb_fut_out_channel*self.future_steps+self.past_steps*self.past_channels+self.future_channels*self.future_steps,hidden_size),
                                                    activation(),
                                                    nn.BatchNorm1d(hidden_size) if use_bn else nn.Dropout(dropout_rate) ,    
                                                    nn.Linear(hidden_size,hidden_size//2), 
                                                    activation(),
                                                    nn.BatchNorm1d(hidden_size//2) if use_bn else nn.Dropout(dropout_rate) ,    
                                                    nn.Linear(hidden_size//2,hidden_size//4),
                                                    activation(),
                                                    nn.BatchNorm1d(hidden_size//4) if use_bn else nn.Dropout(dropout_rate) ,    
                                                    nn.Linear(hidden_size//4,hidden_size//8),
                                                    activation(),
                                                    nn.BatchNorm1d(hidden_size//8) if use_bn else nn.Dropout(dropout_rate) ,    
                                                    nn.Linear(hidden_size//8,self.future_steps*self.mul)))
    def can_be_compiled(self):
        return True                            
    def forward(self, batch):
      
        x =  batch['x_num_past'].to(self.device)
        idx_target = batch['idx_target'][0]
        
        BS = x.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
        
        if self.kind=='nlinear':
            
            x_start = x[:,-1,idx_target].unsqueeze(1)
            x[:,:,idx_target]-=x_start
        
        if self.kind=='alinear':
            x[:,:,idx_target] = 0
        
        if self.kind=='dlinear':
            x_start = x[:,:,idx_target]
            seasonal_init, trend_init = self.decompsition(x_start)
            seasonal_init, trend_init = seasonal_init.permute(0,2,1), trend_init.permute(0,2,1)
            x[:,:,idx_target] = seasonal_init.permute(0,2,1)
            tmp = []
            for j in range(len(self.Linear_Trend)):
               
                tmp.append(self.Linear_Trend[j](trend_init[:,j,:]))

            trend = torch.stack(tmp,2)
            
        if self.simple is False:
            if 'x_num_future' in batch.keys():
                x_future = batch['x_num_future'].to(self.device)
            else:
                x_future = None
                
            tmp = [x,emb_past]
            tot_past = torch.cat(tmp,2).flatten(1)
        


            tmp = [emb_fut]
                          
            if x_future is not None:
                tmp.append(x_future)
           
            tot_future = torch.cat(tmp,2).flatten(1)
            tot = torch.cat([tot_past,tot_future],1)
                
            tot = tot.unsqueeze(2).repeat(1,1,len(self.linear)).permute(0,2,1)
        else:
            tot = x.permute(0,2,1)
        res = []

        for j in range(len(self.linear)):
            res.append(self.linear[j](tot[:,j,:]).reshape(BS,-1,self.mul))
        ## BxLxCxMUL
        res = torch.stack(res,2)

        if self.kind=='nlinear':
            #res BxLxCx3
            #start BxCx1
            res+=x_start.unsqueeze(1)
        

        if self.kind=='dlinear':
            res = res+trend.unsqueeze(3)
        
            
        return res
    


## Copyright https://github.com/yuqinie98/PatchTST/blob/main/LICENSE
## Modified for notation alignmenet and batch structure
## extended to what inside patchtst folder



from torch import  nn
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import List,Union
from ..data_structure.utils import beauty_string
from .utils import  get_scope
from .utils import  get_activation
from .patchtst.layers import series_decomp, PatchTST_backbone
from .utils import Embedding_cat_variables



  
class PatchTST(Base):
    handle_multivariate = True
    handle_future_covariates = False
    handle_categorical_variables = True
    handle_quantile_loss = True
    description = get_scope(handle_multivariate,handle_future_covariates,handle_categorical_variables,handle_quantile_loss)
    
    
    def __init__(self, 
              
                 d_model:int,
                 patch_len:int,
                 kernel_size:int,
                 decomposition:bool=True,
                 activation:str='torch.nn.ReLU',
                 n_head:int=1,
                 n_layer:int=2,
                 stride:int=8,
                 remove_last:bool = False,
                 hidden_size:int=1048,
                 dropout_rate:float=0.1,
                 **kwargs)->None:
        """Initializes the model with specified parameters.https://github.com/yuqinie98/PatchTST/blob/main/
        
        Args:
            d_model (int): The dimensionality of the model.
            patch_len (int): The length of the patches.
            kernel_size (int): The size of the kernel for convolutional layers.
            decomposition (bool, optional): Whether to use decomposition. Defaults to True.
            activation (str, optional): The activation function to use. Defaults to 'torch.nn.ReLU'.
            n_head (int, optional): The number of attention heads. Defaults to 1.
            n_layer (int, optional): The number of layers in the model. Defaults to 2.
            stride (int, optional): The stride for convolutional layers. Defaults to 8.
            remove_last (bool, optional): Whether to remove the last layer. Defaults to False.
            hidden_size (int, optional): The size of the hidden layers. Defaults to 1048.
            dropout_rate (float, optional): The dropout rate for regularization. Defaults to 0.1.
            **kwargs: Additional keyword arguments.
        
        Raises:
            ValueError: If the activation function is not recognized.
        

        """
        super().__init__(**kwargs)

        if activation == 'torch.nn.SELU':
            beauty_string('SELU do not require BN','info',self.verbose)
        if isinstance(activation, str):
            activation = get_activation(activation)
        else:
            beauty_string('There is a bug in pytorch lightening, the constructior is called twice ','info',self.verbose)
        
   
        self.save_hyperparameters(logger=False)
      
        self.remove_last = remove_last
     
        self.emb_past = Embedding_cat_variables(self.past_steps,self.emb_dim,self.embs_past, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        self.emb_fut = Embedding_cat_variables(self.future_steps,self.emb_dim,self.embs_fut, reduction_mode=self.reduction_mode,use_classical_positional_encoder=self.use_classical_positional_encoder,device = self.device)
        emb_past_out_channel = self.emb_past.output_channels
        emb_fut_out_channel = self.emb_fut.output_channels

     
    
        self.past_channels+=emb_past_out_channel
        
    
        # model
        self.decomposition = decomposition
        if self.decomposition:
            self.decomp_module = series_decomp(kernel_size)
            self.model_trend = PatchTST_backbone(c_in=self.past_channels, context_window = self.past_steps, target_window=self.future_steps, patch_len=patch_len, stride=stride, 
                                  max_seq_len=self.past_steps+self.future_steps, n_layers=n_layer, d_model=d_model,
                                  n_heads=n_head, d_k=None, d_v=None, d_ff=hidden_size, norm='BatchNorm', attn_dropout=dropout_rate,
                                  dropout=dropout_rate, act=activation(), key_padding_mask='auto', padding_var=None, 
                                  attn_mask=None, res_attention=True, pre_norm=False, store_attn=False,
                                  pe='zeros', learn_pe=True, fc_dropout=dropout_rate, head_dropout=dropout_rate, padding_patch = 'end',
                                  pretrain_head=False, head_type='flatten', individual=False, revin=True, affine=False,
                                  subtract_last=remove_last, verbose=False)
            self.model_res = PatchTST_backbone(c_in=self.past_channels, context_window = self.past_steps, target_window=self.future_steps, patch_len=patch_len, stride=stride, 
                                  max_seq_len=self.past_steps+self.future_steps, n_layers=n_layer, d_model=d_model,
                                  n_heads=n_head, d_k=None, d_v=None, d_ff=hidden_size, norm='BatchNorm', attn_dropout=dropout_rate,
                                  dropout=dropout_rate, act=activation(), key_padding_mask='auto', padding_var=None, 
                                  attn_mask=None, res_attention=True, pre_norm=False, store_attn=False,
                                  pe='zeros', learn_pe=True, fc_dropout=dropout_rate, head_dropout=dropout_rate, padding_patch = 'end',
                                  pretrain_head=False, head_type='flatten', individual=False, revin=True, affine=False,
                                  subtract_last=remove_last, verbose=False)
        else:
            self.model = PatchTST_backbone(c_in=self.past_channels, context_window = self.past_steps, target_window=self.future_steps, patch_len=patch_len, stride=stride, 
                                  max_seq_len=self.past_steps+self.future_steps, n_layers=n_layer, d_model=d_model,
                                  n_heads=n_head, d_k=None, d_v=None, d_ff=hidden_size, norm='BatchNorm', attn_dropout=dropout_rate,
                                  dropout=dropout_rate, act=activation(), key_padding_mask='auto', padding_var=None, 
                                  attn_mask=None, res_attention=True, pre_norm=False, store_attn=False,
                                  pe='zeros', learn_pe=True, fc_dropout=dropout_rate, head_dropout=dropout_rate, padding_patch = 'end',
                                  pretrain_head=False, head_type='flatten', individual=False, revin=True, affine=False,
                                  subtract_last=remove_last, verbose=False)
    
    
        dim = self.past_channels+emb_fut_out_channel+self.future_channels
        self.final_layer = nn.Sequential(activation(),
                                         nn.Linear(dim, dim*2),
                                         activation(),
                                         nn.Linear(dim*2,self.out_channels*self.mul  ))


    
        #self.final_linear = nn.Sequential(nn.Linear(past_channels,past_channels//2),activation(),nn.Dropout(dropout_rate), nn.Linear(past_channels//2,out_channels)  )
    
    def can_be_compiled(self):
        return True  
    
    def forward(self, batch):           # x: [Batch, Input length, Channel]
        

        x_seq = batch['x_num_past'].to(self.device)#[:,:,idx_target]
        BS = x_seq.shape[0]
        if 'x_cat_future' in batch.keys():
            emb_fut = self.emb_fut(BS,batch['x_cat_future'].to(self.device))
        else:
            emb_fut = self.emb_fut(BS,None)
        if 'x_cat_past' in batch.keys():
            emb_past = self.emb_past(BS,batch['x_cat_past'].to(self.device))
        else:
            emb_past = self.emb_past(BS,None)
            
        tmp_future = [emb_fut]
        if 'x_num_future' in batch.keys():
            x_future = batch['x_num_future'].to(self.device)
            tmp_future.append(x_future)
        
        
        tot = [x_seq,emb_past]
    
        x_seq = torch.cat(tot,axis=2)

        if self.decomposition:
            res_init, trend_init = self.decomp_module(x_seq)
            res_init, trend_init = res_init.permute(0,2,1), trend_init.permute(0,2,1)  # x: [Batch, Channel, Input length]
            res = self.model_res(res_init)
            trend = self.model_trend(trend_init)
            x = res + trend
            x = x.permute(0,2,1)    # x: [Batch, Input length, Channel]
        else:
            x = x_seq.permute(0,2,1)# x: [Batch, Channel, Input length]
            x = self.model(x)
            x = x.permute(0,2,1)    # x: [Batch, Input length, Channel]
        
        
        tmp_future.append(x)
        tmp_future = torch.cat(tmp_future,2)
        output = self.final_layer(tmp_future)
        return output.reshape(BS,self.future_steps,self.out_channels,self.mul)

        
        
        
        


## Copiright https://github.com/PaddlePaddle/PaddleSpatial
## Modified for notation alignmenet,  batch structure 
## extended to what inside d3vae folder

from torch import  nn,optim
import torch

try:
    import lightning.pytorch as pl
    from .base_v2 import Base
    OLD_PL = False
except:
    import pytorch_lightning as pl
    OLD_PL = True
    from .base import Base
from typing import Union
from .d3vae.model import diffusion_generate, denoise_net,pred_net
from .utils import Embedding_cat_variables


from torch.optim.lr_scheduler import StepLR

def copy_parameters(
    net_source: torch.nn.Module,
    net_dest: torch.nn.Module,
    strict= True,
) -> None:
    """
    Copies parameters from one network to another.
    Parameters
    ----------
    net_source
        Input network.
    net_dest
        Output network.
    strict:
        whether to strictly enforce that the keys
        in :attr:`state_dict` match the keys returned by this module's
        :meth:`~torch.nn.Module.state_dict` function. Default: ``True``
    """

    net_dest.load_state_dict(net_source.state_dict(), strict=strict)


class D3VAE(Base):              
    
    def __init__(self,
                 
                 
                 scale=0.1,
                 hidden_size=64,
                 num_layers=2,
                 dropout_rate=0.1,
                 diff_steps=200,
                 loss_type='kl',
                 beta_end=0.01,
                 beta_schedule='linear',
                 channel_mult = 2,
                 mult=1,
                 num_preprocess_blocks=1,
                 num_preprocess_cells=3,
                 num_channels_enc=16,
                 arch_instance = 'res_mbconv',
                 num_latent_per_group=6,
                 num_channels_dec=16,
                 groups_per_scale=2,
                 num_postprocess_blocks=1,
                 num_postprocess_cells=2,
                 beta_start=0,
                 
                 freq='h',
                
                 **kwargs
                 )->None:
        super().__init__(**kwargs)
        input_dim = self.past_channels
        sequence_length = self.past_steps
        prediction_length = self.future_steps
        target_dim = self.out_channels
        embedding_dimension = self.emb_dim
        ##pytotch lightening stuff
        self.save_hyperparameters(logger=False)
        
        
        
        
        self.gen_net = diffusion_generate(target_dim,embedding_dimension,prediction_length,sequence_length,scale,hidden_size,num_layers,dropout_rate,diff_steps,loss_type,beta_end,beta_schedule, channel_mult,mult,
                 num_preprocess_blocks,num_preprocess_cells,num_channels_enc,arch_instance,num_latent_per_group,num_channels_dec,groups_per_scale,num_postprocess_blocks,num_postprocess_cells).to(self.device)
        
        self.denoise_net = denoise_net(target_dim,embedding_dimension,prediction_length,sequence_length,scale,hidden_size,num_layers,dropout_rate,diff_steps,loss_type,beta_end,beta_schedule, channel_mult,mult,
                 num_preprocess_blocks,num_preprocess_cells,num_channels_enc,arch_instance,num_latent_per_group,num_channels_dec,groups_per_scale,num_postprocess_blocks,num_postprocess_cells,beta_start,input_dim,freq,self.embs_past).to(self.device)
        self.diff_step = diff_steps
        self.pred_net = pred_net(target_dim,embedding_dimension,prediction_length,sequence_length,scale,hidden_size,num_layers,dropout_rate,diff_steps,loss_type,beta_end,beta_schedule, channel_mult,mult,
                 num_preprocess_blocks,num_preprocess_cells,num_channels_enc,arch_instance,num_latent_per_group,num_channels_dec,groups_per_scale,num_postprocess_blocks,num_postprocess_cells,beta_start,input_dim,freq,self.embs_fut).to(self.device)
        #self.embedding = DataEmbedding(input_dim, embedding_dimension, freq,dropout_rate)
        
        self.psi = 0.5
        self.gamma = 0.01
        self.lambda1 = 1.0

    def can_be_compiled(self):
        return False   
    def configure_optimizers(self):
        """
        Each model has optim_config and scheduler_config
        
        :meta private:
        """
        optimizer = optim.Adam(self.denoise_net.parameters(),  **self.optim_config)
        self.lr = self.optim_config['lr']
        if self.scheduler_config is not None:
            scheduler = StepLR(optimizer,**self.scheduler_config)
            return [optimizer], [scheduler]
        else:
            return optimizer
        
        
        
    def training_step(self, batch, batch_idx):
        """
        pythotrch lightening stuff
        
        :meta private:
        """
        sample, y_noisy,recon,loss2,total_c = self(batch)
        ##self.compute_loss(batch,y_hat)
        mse_loss = self.loss(sample, y_noisy)
        loss1 = - torch.mean(torch.sum(recon, dim=[1, 2, 3]))
        loss = loss1*self.psi + loss2*self.lambda1 + mse_loss - self.gamma*total_c
        
        
        return loss
        
    def validation_step(self, batch, batch_idx):
        """
        pythotrch lightening stuff
        
        :meta private:
        """
        
        copy_parameters(self.denoise_net, self.pred_net)
        batch_x = batch['x_num_past'].to(self.device)
        batch_x_mark = batch['x_cat_past'].to(self.device)
        batch_y = batch['y'].to(self.device)

        import pdb
        pdb.set_trace()
        _, out, _, _ = self.pred_net(batch_x, batch_x_mark)
        mse = self.loss(out.squeeze(1), batch_y)
        
        
        return mse

        
        
    def forward(self,batch:dict)->torch.tensor:
        
        B = batch['x_num_past'].shape[0]

        
        t = torch.randint(0, self.diff_step, (B,)).long().to(self.device)
        
        batch_x = batch['x_num_past'].to(self.device)
        x_mark = batch['x_cat_past'].to(self.device)
        batch_y = batch['y'].to(self.device)
        
        output, y_noisy, total_c, _, loss2 = self.denoise_net(batch_x, x_mark, batch_y, t)
        recon = output.log_prob(y_noisy)
        sample = output.sample()
        
        
        return sample,y_noisy,recon,loss2,total_c

   
   
   
    

    
    def inference(self, batch:dict)->torch.tensor:
        """Care here, we need to implement it because for predicting the N-step it will use the prediction at step N-1. TODO fix if because I did not implement the
        know continuous variable presence here

        Args:
            batch (dict): batch of the dataloader

        Returns:
            torch.tensor: result
        """
        copy_parameters(self.denoise_net, self.pred_net)

        batch_x = batch['x_num_past'].float().to(self.device)
        batch_x_mark = batch['x_cat_past'].to(self.device)
        _, out, _, _ = self.pred_net(batch_x, batch_x_mark)
        #import pdb
        #pdb.set_trace()
        return torch.permute(out, (0,2,1,3))


DEFAULT_FREQUENCY_MAPPING = {
    "oov": 0,
    "min": 1,  # minutely
    "2min": 2,
    "5min": 3,
    "10min": 4,
    "15min": 5,
    "30min": 6,
    "h": 7,  # hourly
    "H": 7,  # hourly, for compatibility
    "d": 8,  # daily, for compatibility
    "D": 8,  # daily
    "W": 9,  # weekly
}

TTM_LOW_RESOLUTION_MODELS_MAX_CONTEXT = 512


import torch
import torch.nn as nn

from .attn import TwoStageAttentionLayer
from math import ceil

class SegMerging(nn.Module):
    '''
    Segment Merging Layer.
    The adjacent `win_size' segments in each dimension will be merged into one segment to
    get representation of a coarser scale
    we set win_size = 2 in our paper
    '''
    def __init__(self, d_model, win_size, norm_layer=nn.LayerNorm):
        super().__init__()
        self.d_model = d_model
        self.win_size = win_size
        self.linear_trans = nn.Linear(win_size * d_model, d_model)
        self.norm = norm_layer(win_size * d_model)

    def forward(self, x):
        """
        x: B, ts_d, L, d_model
        """
        batch_size, ts_d, seg_num, d_model = x.shape
        pad_num = seg_num % self.win_size
        #import pdb
        #pdb.set_trace()
        if pad_num != 0: 
            pad_num = self.win_size - pad_num
            x = torch.cat((x, x[:, :, -pad_num:, :]), dim = -2)

        seg_to_merge = []
        for i in range(self.win_size):
            seg_to_merge.append(x[:, :, i::self.win_size, :])
   
        x = torch.cat(seg_to_merge, -1)  # [B, ts_d, seg_num/win_size, win_size*d_model]

        x = self.norm(x)
        x = self.linear_trans(x)

        return x

class scale_block(nn.Module):
    '''
    We can use one segment merging layer followed by multiple TSA layers in each scale
    the parameter `depth' determines the number of TSA layers used in each scale
    We set depth = 1 in the paper
    '''
    def __init__(self, win_size, d_model, n_heads, d_ff, depth, dropout, \
                    seg_num = 10, factor=10):
        super(scale_block, self).__init__()

        if (win_size > 1):
            self.merge_layer = SegMerging(d_model, win_size, nn.LayerNorm)
        else:
            self.merge_layer = None
        
        self.encode_layers = nn.ModuleList()

        for i in range(depth):
            self.encode_layers.append(TwoStageAttentionLayer(seg_num, factor, d_model, n_heads, \
                                                        d_ff, dropout))
    
    def forward(self, x):
        _, ts_dim, _, _ = x.shape

        if self.merge_layer is not None:
            x = self.merge_layer(x)
        
        for layer in self.encode_layers:
            x = layer(x)        
        
        return x

class Encoder(nn.Module):
    '''
    The Encoder of Crossformer.
    '''
    def __init__(self, e_blocks, win_size, d_model, n_heads, d_ff, block_depth, dropout,
                in_seg_num = 10, factor=10):
        super(Encoder, self).__init__()
        self.encode_blocks = nn.ModuleList()

        self.encode_blocks.append(scale_block(1, d_model, n_heads, d_ff, block_depth, dropout,\
                                            in_seg_num, factor))
        for i in range(1, e_blocks):
            self.encode_blocks.append(scale_block(win_size, d_model, n_heads, d_ff, block_depth, dropout,\
                                            ceil(in_seg_num/win_size**i), factor))

    def forward(self, x):
        encode_x = []
        encode_x.append(x)
        
        for block in self.encode_blocks:
            x = block(x)
            encode_x.append(x)

        return encode_x